import numpy as np
import scipy.linalg as scl
import scipy.sparse

from deeptime.numeric import sort_eigs

__all__ = ['bootstrapping_count_matrix', 'bootstrapping_dtrajs', 'twostep_count_matrix', 'rank_decision',
           'oom_components', 'equilibrium_transition_matrix']

from deeptime.util.matrix import submatrix


def bootstrapping_dtrajs(dtrajs, lag, N_full, nbs=10000, active_set=None):
    """
    Perform trajectory based re-sampling.

    Parameters
    ----------
    dtrajs : list of discrete trajectories

    lag : int
        lag time

    N_full : int
        Number of states in discrete trajectories.
    nbs : int, optional
        Number of bootstrapping samples
    active_set : ndarray
        Indices of active set, all count matrices will be restricted
        to active set.

    Returns
    -------
    smean : ndarray(N,)
        mean values of singular values
    sdev : ndarray(N,)
        standard deviations of singular values
    """

    # Get the number of simulations:
    Q = len(dtrajs)
    # Get the number of states in the active set:
    if active_set is not None:
        N = active_set.size
    else:
        N = N_full
    # Build up a matrix of count matrices for each simulation. Size is Q*N^2:
    traj_ind = []
    state1 = []
    state2 = []
    q = 0
    for traj in dtrajs:
        traj_ind.append(q * np.ones(traj[:-lag].size))
        state1.append(traj[:-lag])
        state2.append(traj[lag:])
        q += 1
    traj_inds = np.concatenate(traj_ind)
    pairs = N_full * np.concatenate(state1) + np.concatenate(state2)
    data = np.ones(pairs.size)
    Ct_traj = scipy.sparse.coo_matrix((data, (traj_inds, pairs)), shape=(Q, N_full * N_full))
    Ct_traj = Ct_traj.tocsr()

    # Perform re-sampling:
    svals = np.zeros((nbs, N))
    for s in range(nbs):
        # Choose selection:
        sel = np.random.choice(Q, Q, replace=True)
        # Compute count matrix for selection:
        Ct_sel = Ct_traj[sel, :].sum(axis=0)
        Ct_sel = np.asarray(Ct_sel).reshape((N_full, N_full))
        if active_set is not None:
            Ct_sel = submatrix(Ct_sel, active_set)
        svals[s, :] = scl.svdvals(Ct_sel)
    # Compute mean and uncertainties:
    smean = np.mean(svals, axis=0)
    sdev = np.std(svals, axis=0)

    return smean, sdev


def bootstrapping_count_matrix(Ct, nbs=10000):
    """
    Perform bootstrapping on trajectories to estimate uncertainties for singular values of count matrices.

    Parameters
    ----------
    Ct : csr-matrix
        count matrix of the data.

    nbs : int, optional
        the number of re-samplings to be drawn from dtrajs

    Returns
    -------
    smean : ndarray(N,)
        mean values of singular values
    sdev : ndarray(N,)
        standard deviations of singular values
    """
    # Get the number of states:
    N = Ct.shape[0]
    # Get the number of transition pairs:
    T = Ct.sum()
    # Reshape and normalize the count matrix:
    p = Ct.toarray()
    p = np.reshape(p, (N * N,)).astype(np.float)
    p = p / T
    # Perform the bootstrapping:
    svals = np.zeros((nbs, N))
    for s in range(nbs):
        # Draw sample:
        sel = np.random.multinomial(T, p)
        # Compute the count-matrix:
        sC = np.reshape(sel, (N, N))
        # Compute singular values:
        svals[s, :] = scl.svdvals(sC)
    # Compute mean and uncertainties:
    smean = np.mean(svals, axis=0)
    sdev = np.std(svals, axis=0)

    return smean, sdev


def twostep_count_matrix(dtrajs, lag, N):
    """
    Compute all two-step count matrices from discrete trajectories.

    Parameters
    ----------
    dtrajs : list of discrete trajectories

    lag : int
        the lag time for count matrix estimation
    N : int
        the number of states in the discrete trajectories.

    Returns
    -------
    C2t : sparse csc-matrix (N, N, N)
        two-step count matrices for all states. C2t[:, n, :] is a count matrix for each n

    """
    # List all transition triples:
    rows = []
    cols = []
    states = []
    for dtraj in dtrajs:
        if dtraj.size > 2 * lag:
            rows.append(dtraj[0:-2 * lag])
            states.append(dtraj[lag:-lag])
            cols.append(dtraj[2 * lag:])
    row = np.concatenate(rows)
    col = np.concatenate(cols)
    state = np.concatenate(states)
    data = np.ones(row.size)
    # Transform the rows and cols into a single list with N*+2 possible values:
    pair = N * row + col
    # Estimate sparse matrix:
    C2t = scipy.sparse.coo_matrix((data, (pair, state)), shape=(N * N, N))

    return C2t.tocsc()


def rank_decision(smean, sdev, tol=10.0):
    """
    Rank decision based on uncertainties of singular values.

    Parameters
    ----------
    smean : ndarray(N,)
        mean values of singular values for Ct
    sdev : ndarray(N,)
        standard errors of singular values for Ct
    tol : float, optional default=10.0
        accept singular values with signal-to-noise ratio >= tol_svd

    Returns
    -------
    ind : ndarray(N, dtype=bool)
        indicates which singular values are accepted.

    """
    # Determine signal-to-noise ratios of singular values:
    sratio = smean / sdev
    # Return Boolean array of accepted singular values:
    return sratio >= tol


def oom_components(Ct, C2t, rank_ind=None, lcc=None, tol_one=1e-2):
    """
    Compute OOM components and eigenvalues from count matrices:

    Parameters
    ----------
    Ct : ndarray(N, N)
        count matrix from data
    C2t : sparse csc-matrix (N*N, N)
        two-step count matrix from data for all states, columns enumerate
        intermediate steps.
    rank_ind : ndarray(N, dtype=bool), optional, default=None
        indicates which singular values are accepted. By default, all non-
        zero singular values are accepted.
    lcc : ndarray(N,), optional, default=None
        largest connected set of the count-matrix. Two step count matrix
        will be reduced to this set.
    tol_one : float, optional, default=1e-2
        keep eigenvalues of absolute value less or equal 1+tol_one.

    Returns
    -------
    Xi : ndarray(M, N, M)
        matrix of set-observable operators
    oom_information_state_vector: ndarray(M,)
        information state vector of OOM
    oom_evaluator : ndarray(M,)
        evaluator of OOM
    l : ndarray(M,)
        eigenvalues from OOM
    """
    import deeptime.markov.tools.estimation as me
    # Decompose count matrix by SVD:
    if lcc is not None:
        Ct_svd = me.largest_connected_submatrix(Ct, lcc=lcc)
        N1 = Ct.shape[0]
    else:
        Ct_svd = Ct
    V, s, W = scl.svd(Ct_svd, full_matrices=False)
    # Make rank decision:
    if rank_ind is None:
        ind = (s >= np.finfo(float).eps)
    V = V[:, rank_ind]
    s = s[rank_ind]
    W = W[rank_ind, :].T

    # Compute transformations:
    F1 = np.dot(V, np.diag(s ** -0.5))
    F2 = np.dot(W, np.diag(s ** -0.5))

    # Apply the transformations to C2t:
    N = Ct_svd.shape[0]
    M = F1.shape[1]
    Xi = np.zeros((M, N, M))
    for n in range(N):
        if lcc is not None:
            C2t_n = C2t[:, lcc[n]]
            C2t_n = _reshape_sparse(C2t_n, (N1, N1))
            C2t_n = me.largest_connected_submatrix(C2t_n, lcc=lcc)
        else:
            C2t_n = C2t[:, n]
            C2t_n = _reshape_sparse(C2t_n, (N, N))
        Xi[:, n, :] = np.dot(F1.T, C2t_n.dot(F2))

    # Compute sigma:
    c = np.sum(Ct_svd, axis=1)
    sigma = np.dot(F1.T, c)
    # Compute eigenvalues:
    Xi_S = np.sum(Xi, axis=1)
    eigenvalues, right_eigenvectors = scl.eig(Xi_S.T)
    # Restrict eigenvalues to reasonable range:
    ind = np.where(np.logical_and(np.abs(eigenvalues) <= (1 + tol_one), np.real(eigenvalues) >= 0.0))[0]
    eigenvalues = eigenvalues[ind]
    right_eigenvectors = right_eigenvectors[:, ind]
    # Sort and extract omega
    eigenvalues, right_eigenvectors = sort_eigs(eigenvalues, right_eigenvectors)
    omega = np.real(right_eigenvectors[:, 0])
    omega = omega / np.dot(omega, sigma)

    return Xi, omega, sigma, eigenvalues


def equilibrium_transition_matrix(Xi, omega, sigma, reversible=True, return_lcc=True):
    """
    Compute equilibrium transition matrix from OOM components:

    Parameters
    ----------
    Xi : ndarray(M, N, M)
        matrix of set-observable operators
    omega: ndarray(M,)
        information state vector of OOM
    sigma : ndarray(M,)
        evaluator of OOM
    reversible : bool, optional, default=True
        symmetrize corrected count matrix in order to obtain
        a reversible transition matrix.
    return_lcc: bool, optional, default=True
        return indices of largest connected set.

    Returns
    -------
    Tt_Eq : ndarray(N, N)
        equilibrium transition matrix
    lcc : ndarray(M,)
        the largest connected set of the transition matrix.
    """
    import deeptime.markov.tools.estimation as me

    # Compute equilibrium transition matrix:
    Ct_Eq = np.einsum('j,jkl,lmn,n->km', omega, Xi, Xi, sigma)
    # Remove negative entries:
    Ct_Eq[Ct_Eq < 0.0] = 0.0
    # Compute transition matrix after symmetrization:
    pi_r = np.sum(Ct_Eq, axis=1)
    if reversible:
        pi_c = np.sum(Ct_Eq, axis=0)
        pi_sym = pi_r + pi_c
        # Avoid zero row-sums. States with zero row-sums will be eliminated by active set update.
        ind0 = np.where(pi_sym == 0.0)[0]
        pi_sym[ind0] = 1.0
        Tt_Eq = (Ct_Eq + Ct_Eq.T) / pi_sym[:, None]
    else:
        # Avoid zero row-sums. States with zero row-sums will be eliminated by active set update.
        ind0 = np.where(pi_r == 0.0)[0]
        pi_r[ind0] = 1.0
        Tt_Eq = Ct_Eq / pi_r[:, None]

    # Perform active set update:
    lcc = me.largest_connected_set(Tt_Eq)
    Tt_Eq = me.largest_connected_submatrix(Tt_Eq, lcc=lcc)

    if return_lcc:
        return Tt_Eq, lcc
    else:
        return Tt_Eq


def _reshape_sparse(A, shape):
    nrows, ncols = A.shape
    if nrows * ncols != shape[0] * shape[1]:
        raise ValueError("Matrix dimensions must agree.")
    rows, cols = A.nonzero()
    flat_indices = rows * ncols + cols
    newrows, newcols = divmod(flat_indices, shape[1])
    data = np.array(A[A.nonzero()].tolist()).flatten()
    Anew = scipy.sparse.csc_matrix((data, (newrows, newcols)), shape=shape)

    return Anew



def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration

    config = Configuration('data', parent_package, top_path)
    config.add_data_files(
        'data/double_well_discrete.npz',
    )
    config.add_extension('_data_bindings',
                         sources=['src/data_module.cpp'],
                         include_dirs=['include'],
                         language='c++',
                         )
    return config



import numpy as np

from ._systems import TimeDependentSystem
from ..util.data import TimeLaggedDataset
from ..util.decorators import plotting_function


class BickleyJet(TimeDependentSystem):
    r"""Implementation of the Bickley jet. Based on :footcite:`hadjighasem2016spectral`.

    The parameters are set to

    .. math::

        \begin{aligned}
            U_0 &= 5.4138 \times \frac{10^6\mathrm{m}}{\mathrm{day}},\\
            L_0 &= 1.77 \times 10^6\,\mathrm{m},\\
            r_0 &= 6.371 \times 10^6\,\mathrm{m},\\
            c &= (0.1446, 0.205, 0.461)^\top U_0,\\
            \mathrm{eps} &= (0.075, 0.15, 0.3)^\top,\\
            k &= (2,4,6)^\top \frac{1}{r_0}.
        \end{aligned}

    Parameters
    ----------
    h : float
        Step size of the RK45 integrator.
    n_steps : int
        Number of steps before recording it in the trajectory.
    full_periodic : bool, optional, default=True
        Whether the quasi-periodic boundary is already built into right-hand side evaluation.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, h: float, n_steps: int, full_periodic=True):
        from ._data_bindings import BickleyJetFullPeriodic as BickleyJetImplFP
        from ._data_bindings import BickleyJetPartiallyPeriodic as BickleyJetImplPP
        super().__init__(BickleyJetImplFP() if full_periodic else BickleyJetImplPP(), h, n_steps)
        self.full_periodic = full_periodic
        self.periodic_bc = True

    @property
    def U0(self):
        r""" The characteristic velocity scale :math:`U_0 = 5.4138` in :math:`10^6 \frac{\mathrm{m}}{\mathrm{day}}`. """
        return self._impl.U0
    
    @property
    def L0(self):
        r""" The characteristic length scale :math:`L_0 = 1.77` in :math:`10^6\;\mathrm{m}`. """
        return self._impl.L0

    @property
    def r0(self):
        r""" The mean radius of the earth :math:`r_0 = 6.371` in :math:`10^6\;\mathrm{m}`. """
        return self._impl.r0

    @property
    def c(self):
        r""" Traveling speeds :math:`c = (0.1446, 0.205, 0.461)^\top U_0`. """
        return self._impl.c

    @property
    def eps(self):
        r""" Wave amplitudes :math:`\varepsilon = (0.075, 0.15, 0.3)^\top`. """
        return self._impl.eps

    @property
    def k(self):
        r""" Wave numbers :math:`k = (2,4,6)^\top \frac{1}{r_0}`. """
        return self._impl.k

    @property
    def periodic_bc(self) -> bool:
        r""" Whether periodic boundary conditions are applied.

        :getter: Yields the current value. True corresponds to periodic boundaries.
        :setter: Sets a new value.
        :type: bool
        """
        return self._periodic

    @periodic_bc.setter
    def periodic_bc(self, value: bool):
        if self.full_periodic and not value:
            raise RuntimeError("The system respects boundary conditions fully, please use full_periodic=False.")
        self._periodic = value

    @staticmethod
    def apply_periodic_boundary_conditions(trajectory, inplace=False):
        r""" Applies periodic boundary conditions for domain :math:`\Omega = [0, 20)\times [-3, 3)`.

        Notes
        -----
        This method operates not in-place by default, i.e., makes a copy of the trajectory. This behavior
        can be changed by setting `inplace=True`.

        Parameters
        ----------
        trajectory : (..., 2) ndarray
            Input trajectory, last axis must have length two for x and y, respectively.
        inplace : bool, optional, default=False
            Whether to operate in-place.

        Returns
        -------
        trajectory
            Trajectory with applied boundary conditions.
        """
        if not inplace:
            trajectory = trajectory.copy()
        trajectory[..., 0] = np.mod(trajectory[..., 0], 20)  # periodicity in x direction
        return trajectory

    def trajectory(self, t0, x0, length, seed=-1, n_jobs=None, return_time=False):
        r""" Generates one or multiple trajectories for the Bickley jet.

        Parameters
        ----------
        t0 : array_like
            The initial time. Can be picked as single float across all test points or individually.
        x0 : array_like
            The initial condition. Must be compatible in shape to a (n_test_points, dimension)-array.
        length : int
            The length of the trajectory that is to be generated.
        seed : int, optional, default=-1
            The random seed. In case it is specified to be something else than `-1`, n_jobs must be set to `n_jobs=1`.
        n_jobs : int, optional, default=None
            Specify number of jobs according to :meth:`deeptime.util.parallel.handle_n_jobs`.
        return_time : bool, optional, default=False
            Whether to return the evaluation times too.
        """
        tarr, traj = super().trajectory(t0, x0, length, seed, n_jobs, return_time=True)
        if self.periodic_bc and not self.full_periodic:
            BickleyJet.apply_periodic_boundary_conditions(traj, inplace=True)
        return traj if not return_time else (tarr, traj)

    def generate(self, n_particles, n_jobs=None, seed=None) -> np.ndarray:
        """Generates a trajectory with a fixed number of particles / test points for 401 evaluation steps, i.e.,
        `401 * self.n_steps * self.h` integration steps.

        Parameters
        ----------
        n_particles : int
            Number of particles.
        n_jobs : int, optional, default=None
            Number of jobs.
        seed : int or None, optional, default=None
            Random seed used for initialization of particle positions at :math:`t=0`.

        Returns
        -------
        Z : np.ndarray (m, 401, 2)
            Trajectories for m uniformly distributed test points in Omega = [0, 20] x [-3, 3].
        """
        state = np.random.RandomState(seed)
        X = np.vstack((state.uniform(0, 20, (n_particles,)), state.uniform(-3, 3, (n_particles, ))))
        return self.trajectory(0, X.T, 401, n_jobs=n_jobs, return_time=False)

    @staticmethod
    def to_3d(data: np.ndarray, radius: float = 1.) -> np.ndarray:
        r"""Maps a generated trajectory into 3d space by transforming it through

        .. math::

            \begin{pmatrix} x \\ y \end{pmatrix} \mapsto \begin{pmatrix}
                r\cdot \cos\left( 2\pi \frac{x}{20} \right) \\
                r\cdot \sin\left( 2\pi \frac{x}{20} \right) \\
                \frac{y}{3}
            \end{pmatrix}

        which means that the particles are now on the surface of a cylinder with a fixed radius, i.e.,
        periodic boundary conditions are directly encoded in the space.

        Parameters
        ----------
        data : (T, 2) ndarray
            The generated trajectory.
        radius : float, default=1
            The radius of the cylinder.
        Returns
        -------
        xyz : (T, 3) np.ndarray
            Trajectory on the cylinder.
        """
        t = data[..., 0] * 2. * np.pi / 20.
        z = data[..., 1]
        x_3d = radius * np.cos(t)
        y_3d = radius * np.sin(t)
        xyz = np.stack((x_3d, y_3d, z / 3))
        return xyz.T


class BickleyJetDataset:

    def __init__(self, trajectory):
        self.data = trajectory

    def __len__(self):
        return len(self.data)

    def __getitem__(self, item):
        return self.data[item]

    @plotting_function
    def make_animation(self, **kw):  # pragma: no cover
        import matplotlib as mpl
        import matplotlib.pyplot as plt
        import matplotlib.animation as animation

        agg_backend = kw.get('agg_backend', True)
        figsize = kw.get("figsize", (8, 6))
        title = kw.get("title", None)
        s = kw.get("s", None)
        c = kw.get("c", None)
        stride = kw.get("stride", 1)
        cmap = kw.get('cmap', 'jet')
        edgecolor = kw.get('edgecolor', 'k')
        repeat = kw.get('repeat', False)
        interval = kw.get('interval', 50)
        fig = kw.get('fig', None)
        ax = kw.get('ax', None)
        max_frame = kw.get('max_frame', None)

        data = self.data[:max_frame:stride]
        n_particles = data.shape[1]

        if agg_backend:
            backend_ = mpl.get_backend()
            mpl.use("Agg")  # Prevent showing stuff

        if s is None:
            s = np.empty((n_particles,))
            s.fill(300)
        if c is None:
            c = np.empty((n_particles,))
            c.fill(0.5)

        if fig is None or ax is None:
            fig, ax = plt.subplots(1, 1, figsize=figsize)
        ax.set_title(title)
        ax.set_xlim([0, 20])
        ax.set_ylim([-3, 3])

        plot_handles = []
        s_handle = ax.scatter(*data[0].T, s=s, c=c, vmin=0, vmax=1,
                              cmap=cmap, edgecolor=edgecolor)
        plot_handles.append(s_handle)

        def update(i):
            handle = plot_handles[0]
            handle.set_offsets(data[::stride][i])
            return plot_handles

        ani = animation.FuncAnimation(fig, update, interval=interval, blit=True, repeat=repeat,
                                      frames=data[::stride].shape[0])

        if agg_backend:
            mpl.use(backend_)  # Reset backend

        return ani

    def endpoints_dataset(self):
        return BickleyJetEndpointsDataset(self.data[0], self.data[-1])


class BickleyJetEndpointsDataset(TimeLaggedDataset):

    def __init__(self, data_0, data_t):
        super().__init__(data_0, data_t)

    def to_3d(self, radius: float = 1.):
        r""" See :meth:`BickleyJet.to_3d`. """
        return BickleyJetEndpointsDataset3D(
            BickleyJet.to_3d(self.data, radius), BickleyJet.to_3d(self.data_lagged, radius)
        )


class BickleyJetEndpointsDataset3D(TimeLaggedDataset):

    def __init__(self, data: np.ndarray, data_lagged: np.ndarray):
        super().__init__(data, data_lagged)
        assert data.shape[1] == 3
        assert data_lagged.shape[1] == 3

    def cluster(self, n_bins):
        from deeptime.clustering import BoxDiscretization

        disc = BoxDiscretization(3, n_bins).fit(np.concatenate((self.data, self.data_lagged))).fetch_model()
        traj1 = disc.transform_onehot(self.data)
        traj2 = disc.transform_onehot(self.data_lagged)

        return BickleyJetEndpointsDataset3DClustered(traj1, traj2)


class BickleyJetEndpointsDataset3DClustered(TimeLaggedDataset):

    def __init__(self, data, data_lagged):
        super().__init__(data, data_lagged)



import numpy as np


class Kernel:
    r""" The base class of all kernels. Provides interfaces to evaluating the kernel on points in state space
    as well as computing the kernel matrix / Gramian.
    """

    def __call__(self, x, y):
        r""" Computes the value of the kernel at two specific points.

        Parameters
        ----------
        x : (d,) ndarray
            X point.
        y : (d,) ndarray
            Y point.

        Returns
        -------
        kxy : float
            The kernel evaluation :math:`\kappa (x, y)`.

        Notes
        -----
        This dispatches to :meth:`_evaluate` which is an interface method intended to be overridden
        by a specific kernel implementation.
        """
        return self._evaluate(x, y)

    def _evaluate(self, x, y) -> float:
        raise NotImplementedError("Not implemented, please provide a subclass!")

    def gram(self, data: np.ndarray) -> np.ndarray:
        r""" Computes the corresponding Gram matrix, see also :meth:`apply`.

        Parameters
        ----------
        data : (T, d) ndarray
            The data array.

        Returns
        -------
        G : (T, T) ndarray
            The kernel Gramian with :code:`G[i, j] = k(x[i], x[j])`.
        """
        return self.apply(data_1=data, data_2=data)

    def apply(self, data_1: np.ndarray, data_2: np.ndarray) -> np.ndarray:
        r""" Applies the kernel to data arrays.

        Given :math:`x\in\mathbb{R}^{T_1 \times d}` and
        :math:`y\in\mathbb{R}^{T_2\times d}`, it yields a kernel matrix

        .. math::

            K = (\kappa(x_i, y_j))_{i,j}\in\mathbb{R}^{T_1\times T_2}.

        Note that this corresponds to the kernel Gramian in case :math:`x = y`.

        Parameters
        ----------
        data_1 : (T_1, d) ndarray
            Data array.
        data_2 : (T_2, d) ndarray
            Data array.

        Returns
        -------
        K : (T_1, T_2) ndarray
            The kernel matrix.
        """
        m = data_1.shape[0]
        n = data_2.shape[0]

        result = np.empty([m, n], dtype=data_1.dtype)
        for i in range(m):
            for j in range(n):
                result[i, j] = self(data_1[i], data_2[j])
        return result

    def __mul__(self, other):
        r""" Multiplication with another kernel.

        Parameters
        ----------
        other : Kernel
            The other kernel.

        Returns
        -------
        result : ProductKernel
            Product of this and the other kernel.
        """
        if not isinstance(other, Kernel):
            raise ValueError("Kernel can only be multiplied with other kernels.")
        return ProductKernel(self, other)

    __rmul__ = __mul__  # no difference here


class ProductKernel(Kernel):

    def __init__(self, k1: Kernel, k2: Kernel):
        self._k1 = k1
        self._k2 = k2

    def _evaluate(self, x, y) -> float:
        return self._k1(x, y) * self._k2(x, y)

    def __str__(self):
        return f"{self._k1} * {self._k2}"


def is_torch_kernel(kernel: Kernel) -> bool:
    r""" Tests whether given kernel supports PyTorch. Note that even if the kernel does not support PyTorch it
    (depending on the case) might still be usable in conjunction with PyTorch code at a performance penalty, e.g.,
    when no gradients are required for the kernel evaluation.

    Parameters
    ----------
    kernel : Kernel
        A kernel instance.

    Returns
    -------
    is_torch_kernel : bool
        True if it supports PyTorch, False otherwise.
    """
    return hasattr(kernel, 'apply_torch')



"""This module contains function for the Transition Path Theory (TPT)
analysis of Markov models.

__moduleauthor__ = "Benjamin Trendelkamp-Schroer, Frank Noe, Martin Scherer"

"""
from typing import Optional, Iterable

import numpy as np

from .tools import flux as tptapi

from deeptime.base import Model
from deeptime.markov.tools.flux import to_netflux, pathways, coarsegrain

from ..util.types import ensure_array


class ReactiveFlux(Model):
    r""" The A->B reactive flux from transition path theory (TPT).

    This object describes a reactive flux, i.e. a network of fluxes from a set of source states A, to a set of
    sink states B, via a set of intermediate nodes. Every node has three properties: the stationary probability mu,
    the forward committor qplus and the backward committor qminus. Every pair of edges has the following properties:
    a flux, generally a net flux that has no unnecessary back-fluxes, and optionally a gross flux.

    Flux objects can be used to compute transition pathways (and their weights) from A to B, the total flux, the
    total transition rate or mean first passage time, and they can be coarse-grained onto a set discretization
    of the node set.

    Fluxes can be computed using transition path theory - see :footcite:`metzner2009transition`
    and :func:`deeptime.markov.tools.tpt`.

    Parameters
    ----------
    source_states : array_like
        List of integer state labels for set A
    target_states : array_like
        List of integer state labels for set B
    net_flux : (n,n) ndarray or scipy sparse matrix
        effective or net flux of A->B pathways
    stationary_distribution : (n,) ndarray (optional)
        Stationary vector
    qminus : (n,) ndarray (optional)
        Backward committor for A->B reaction
    qplus : (n,) ndarray (optional)
        Forward committor for A-> B reaction
    gross_flux : (n,n) ndarray or scipy sparse matrix
        gross flux of A->B pathways, if available

    Notes
    -----
    Reactive flux contains a flux network from educt states (A) to product states (B).

    See also
    --------
    reactive_flux : Method that produces ReactiveFlux instances
    deeptime.markov.msm.MarkovStateModel.reactive_flux : TPT analysis based on a Markov state model

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, source_states, target_states, net_flux, stationary_distribution=None,
                 qminus=None, qplus=None, gross_flux=None):
        # set data
        super().__init__()
        self._source_states = source_states
        self._target_states = target_states
        self._net_flux = net_flux
        self._stationary_distribution = stationary_distribution
        self._qminus = qminus
        self._qplus = qplus
        self._gross_flux = gross_flux
        # compute derived quantities:
        self._totalflux = tptapi.total_flux(net_flux, source_states)
        self._kAB = tptapi.rate(self._totalflux, stationary_distribution, qminus)

    @property
    def n_states(self):
        """number of states."""
        return self._net_flux.shape[0]

    @property
    def source_states(self):
        """set of reactant (source) states."""
        return self._source_states

    @property
    def target_states(self):
        """set of product (target) states"""
        return self._target_states

    @property
    def intermediate_states(self):
        """set of intermediate states"""
        return list(set(range(self.n_states)) - set(self._source_states) - set(self._target_states))

    @property
    def stationary_distribution(self):
        """stationary distribution"""
        return self._stationary_distribution

    @property
    def net_flux(self):
        r""" Effective or net flux. Units are :math:`1/ \mathrm{time}`. """
        return self._net_flux

    @property
    def gross_flux(self):
        r""" Gross :math:`A\\rightarrow B` flux. Units are :math:`1/ \mathrm{time}`. """
        return self._gross_flux

    @property
    def forward_committor(self):
        """forward committor probability"""
        return self._qplus

    @property
    def backward_committor(self):
        """backward committor probability"""
        return self._qminus

    @property
    def total_flux(self):
        r""" The total flux. Units are :math:`1/ \mathrm{time}`. """
        return self._totalflux

    @property
    def rate(self):
        r""" Rate (inverse mfpt) of :math:`A\\rightarrow B` transitions in units of :math:`1/ \mathrm{time}`. """
        return self._kAB

    @property
    def mfpt(self):
        """ Mean-first-passage-time (inverse rate) of :math:`A\\rightarrow B` transitions. """
        return 1. / self._kAB

    def pathways(self, fraction=1.0, maxiter=1000):
        """Decompose flux network into dominant reaction paths.

        Parameters
        ----------
        fraction : float, optional
            Fraction of total flux to assemble in pathway decomposition
        maxiter : int, optional
            Maximum number of pathways for decomposition

        Returns
        -------
        paths : list
            List of dominant reaction pathways
        capacities: list
            List of capacities corresponding to each reactions pathway in paths
        """
        return pathways(self._net_flux, self.source_states, self.target_states,
                        fraction=fraction, maxiter=maxiter)

    @staticmethod
    def _pathways_to_flux(paths, pathfluxes, n=None):
        """Sums up the flux from the pathways given

        Parameters
        -----------
        paths : list of int-arrays
        list of pathways

        pathfluxes : double-array
            array with path fluxes

        n : int or None
            number of states. If not set, will be automatically determined.

        Returns
        -------
        flux : (n,n) ndarray of float
            the flux containing the summed path fluxes

        """
        if n is None:
            n = 0
            for p in paths:
                n = max(n, np.max(p))
            n += 1

        # initialize flux
        flux = np.zeros((n, n))
        for i, p in enumerate(paths):
            for t in range(len(p) - 1):
                flux[p[t], p[t + 1]] += pathfluxes[i]
        return flux

    def major_flux(self, fraction=0.9):
        """ Returns the main pathway part of the net flux comprising at most the requested fraction of the full flux.
        """
        paths, pathfluxes = self.pathways(fraction=fraction)
        return self._pathways_to_flux(paths, pathfluxes, n=self.n_states)

    # this will be a private function in tpt. only Parameter left will be the sets to be distinguished
    def _compute_coarse_sets(self, user_sets):
        """Computes the sets to coarse-grain the tpt flux to.

        Parameters
        ----------
        user_sets : list of int-iterables
            sets of states that shall be distinguished in the coarse-grained flux.

        Returns
        -------
        sets : list of int-iterables
            sets to compute tpt on. These sets still respect the boundary between
            A, B and the intermediate tpt states.

        Notes
        -----
        Given the sets that the user wants to distinguish, the
        algorithm will create additional sets if necessary

           * If states are missing in user_sets, they will be put into a
            separate set
           * If sets in user_sets are crossing the boundary between A, B and the
             intermediates, they will be split at these boundaries. Thus each
             set in user_sets can remain intact or be split into two or three
             subsets

        """
        # set-ify everything
        setA = set(self.source_states)
        setB = set(self.target_states)
        setI = set(self.intermediate_states)
        raw_sets = [set(user_set) for user_set in user_sets]

        # anything missing? Compute all listed states
        set_all = set(range(self.n_states))
        set_all_user = []
        for user_set in raw_sets:
            set_all_user += user_set
        set_all_user = set(set_all_user)
        # ... and add all the unlisted states in a separate set
        set_rest = set_all - set_all_user
        if len(set_rest) > 0:
            raw_sets.append(set_rest)

        # split sets
        source_sets = []
        intermediate_sets = []
        target_sets = []
        for raw_set in raw_sets:
            s = raw_set.intersection(setA)
            if len(s) > 0:
                source_sets.append(s)
            s = raw_set.intersection(setI)
            if len(s) > 0:
                intermediate_sets.append(s)
            s = raw_set.intersection(setB)
            if len(s) > 0:
                target_sets.append(s)
        tpt_sets = source_sets + intermediate_sets + target_sets
        source_indexes = list(range(0, len(source_sets)))
        target_indexes = list(range(len(source_sets) + len(intermediate_sets), len(tpt_sets)))

        return tpt_sets, source_indexes, target_indexes

    def coarse_grain(self, user_sets):
        """Coarse-grains the flux onto user-defined sets.

        Parameters
        ----------
        user_sets : list of int-iterables
            sets of states that shall be distinguished in the coarse-grained flux.

        Returns
        -------
        (sets, tpt) : (list of int-iterables, ReactiveFlux)
            sets contains the sets tpt is computed on. The tpt states of the new
            tpt object correspond to these sets of states in this order. Sets might
            be identical, if the user has already provided a complete partition that
            respects the boundary between A, B and the intermediates. If not, Sets
            will have more members than provided by the user, containing the
            "remainder" states and reflecting the splitting at the A and B
            boundaries.
            tpt contains a new tpt object for the coarse-grained flux. All its
            quantities (gross_flux, net_flux, A, B, committor, backward_committor)
            are coarse-grained to sets.

        Notes
        -----
        All user-specified sets will be split (if necessary) to
        preserve the boundary between A, B and the intermediate
        states.

        """
        # coarse-grain sets
        tpt_sets, source_indices, target_indices = self._compute_coarse_sets(user_sets)
        nnew = len(tpt_sets)

        # coarse-grain flux
        # Here we should branch between sparse and dense implementations, but currently there is only a dense version.
        flux_coarse = coarsegrain(self._gross_flux, tpt_sets)
        net_flux_coarse = to_netflux(flux_coarse)

        # coarse-grain stationary probability and committors - this can be done all dense
        pstat_coarse = np.zeros(nnew)
        forward_committor_coarse = np.zeros(nnew)
        backward_committor_coarse = np.zeros(nnew)
        for i in range(0, nnew):
            I = list(tpt_sets[i])
            statdist_subselection = self._stationary_distribution[I]
            pstat_coarse[i] = np.sum(statdist_subselection)
            # normalized stationary probability over I
            statdist_subselection_prob = statdist_subselection / pstat_coarse[i]
            forward_committor_coarse[i] = np.dot(statdist_subselection_prob, self._qplus[I])
            backward_committor_coarse[i] = np.dot(statdist_subselection_prob, self._qminus[I])

        res = ReactiveFlux(source_indices, target_indices, net_flux=net_flux_coarse,
                           stationary_distribution=pstat_coarse, qminus=backward_committor_coarse,
                           qplus=forward_committor_coarse, gross_flux=flux_coarse)
        return tpt_sets, res


def reactive_flux(transition_matrix: np.ndarray, source_states: Iterable[int], target_states: Iterable[int],
                  stationary_distribution=None, qminus=None, qplus=None,
                  transition_matrix_tolerance: Optional[float] = None) -> ReactiveFlux:
    r""" Computes the A->B reactive flux using transition path theory (TPT).

    Parameters
    ----------
    transition_matrix : (M, M) ndarray or scipy.sparse matrix
        The transition matrix.
    source_states : array_like
        List of integer state labels for set A
    target_states : array_like
        List of integer state labels for set B
    stationary_distribution : (M,) ndarray, optional, default=None
        Stationary vector. If None is computed from the transition matrix internally.
    qminus : (M,) ndarray (optional)
        Backward committor for A->B reaction
    qplus : (M,) ndarray (optional)
        Forward committor for A-> B reaction
    transition_matrix_tolerance : float, optional, default=None
        Tolerance with which is checked whether the input is actually a transition matrix. If None (default),
        no check is performed.

    Returns
    -------
    tpt: deeptime.markov.tools.flux.ReactiveFlux object
        A python object containing the reactive A->B flux network
        and several additional quantities, such as stationary probability,
        committors and set definitions.

    Notes
    -----
    The central object used in transition path theory is the forward and backward comittor function.

    TPT (originally introduced in :footcite:`weinan2006towards`) for continous systems has a
    discrete version outlined in :footcite:`metzner2009transition`. Here, we use the transition
    matrix formulation described in :footcite:`noe2009constructing`.

    See also
    --------
    ReactiveFlux

    References
    ----------
    .. footbibliography::
    """
    import deeptime.markov.tools.analysis as msmana

    source_states = ensure_array(source_states, dtype=int)
    target_states = ensure_array(target_states, dtype=int)

    if len(source_states) == 0 or len(target_states) == 0:
        raise ValueError('set A or B is empty')

    n_states = transition_matrix.shape[0]
    if len(source_states) > n_states or len(target_states) > n_states \
            or max(source_states) > n_states or max(target_states) > n_states:
        raise ValueError('set A or B defines more states than the given transition matrix.')

    if transition_matrix_tolerance is not None and \
            msmana.is_transition_matrix(transition_matrix, tol=transition_matrix_tolerance):
        raise ValueError('given matrix T is not a transition matrix')

    # we can compute the following properties from either dense or sparse T
    # stationary dist
    if stationary_distribution is None:
        stationary_distribution = msmana.stationary_distribution(transition_matrix)
    # forward committor
    if qplus is None:
        qplus = msmana.committor(transition_matrix, source_states, target_states, forward=True)
    # backward committor
    if qminus is None:
        if msmana.is_reversible(transition_matrix, mu=stationary_distribution):
            qminus = 1.0 - qplus
        else:
            qminus = msmana.committor(transition_matrix, source_states, target_states, forward=False,
                                      mu=stationary_distribution)
    # gross flux
    grossflux = tptapi.flux_matrix(transition_matrix, stationary_distribution, qminus, qplus, netflux=False)
    # net flux
    netflux = to_netflux(grossflux)

    # construct flux object
    return ReactiveFlux(source_states, target_states, net_flux=netflux,
                        stationary_distribution=stationary_distribution,
                        qminus=qminus, qplus=qplus, gross_flux=grossflux)



from typing import Callable, Union, List, Optional

import numpy as np
import torch


class DLEstimatorMixin:
    r""" Estimator subclass which offers some deep-learning estimators commonly used functionality.
    """

    @property
    def learning_rate(self):
        r""" Sets or yields a learning rate. Note that :meth:`setup_optimizer` should be called on update to propagate
        the changes.

        :type: float
        """
        if not hasattr(self, '_learning_rate'):
            self._learning_rate = 5e-4
        return self._learning_rate

    @learning_rate.setter
    def learning_rate(self, value):
        self._learning_rate = value

    @property
    def device(self):
        r""" The device on which the estimator's PyTorch module(s) are operating. """
        if not hasattr(self, '_device'):
            self._device = None
        return self._device

    @device.setter
    def device(self, value):
        self._device = value

    @property
    def dtype(self):
        r""" The data type under which the estimator operates.

        :getter: Gets the currently set data type.
        :setter: Sets a new data type, must be one of np.float32, np.float64
        :type: numpy data type type
        """
        if not hasattr(self, '_dtype'):
            self._dtype = np.float32
        return self._dtype

    @dtype.setter
    def dtype(self, value):
        assert value in (np.float32, np.float64), "only float32 and float64 are supported."
        self._dtype = value

    @property
    def optimizer(self) -> Optional[torch.optim.Optimizer]:
        r""" The optimizer that is used.

        :getter: Gets the currently configured optimizer.
        :setter: Sets a new optimizer based on optimizer name (string) or optimizer class (class reference).
        :type: torch.optim.Optimizer
        """
        if not hasattr(self, '_optimizer'):
            self._optimizer = None
        return self._optimizer

    def setup_optimizer(self, kind: Union[str, Callable], parameters: List):
        r""" Initializes a new optimizer on a list of parameters.

        Parameters
        ----------
        kind : str or Callable
            The optimizer.
        parameters : list of parameters
            The parameters
        """
        unique_params = list(set(parameters))
        if isinstance(kind, str):
            known_optimizers = {'Adam': torch.optim.Adam, 'SGD': torch.optim.SGD, 'RMSprop': torch.optim.RMSprop}
            if kind not in known_optimizers.keys():
                raise ValueError(f"Unknown optimizer type, supported types are {known_optimizers.keys()}. "
                                 f"If desired, you can also pass the class of the "
                                 f"desired optimizer rather than its name.")
            kind = known_optimizers[kind]
        self._optimizer = kind(params=unique_params, lr=self.learning_rate)



import itertools
import numbers
from typing import Tuple, Optional, Union, List

import numpy as np
from scipy.sparse import spmatrix, issparse

from .data import TimeLaggedDataset, TrajectoryDataset, TrajectoriesDataset


def atleast_nd(ary, ndim, pos=0):
    r"""View input as array with at least `ndim` dimensions.
    New unit dimensions are inserted at the index given by `pos` if
    necessary.

    Parameters
    ----------
    ary : array_like
        The input array. Non-array inputs are converted to arrays.
        Arrays that already have `ndim` or more dimensions are
        preserved.
    ndim : scalar
        The minimum number of dimensions required.
    pos : int, optional
        The index to insert the new dimensions. May range from
        ``-ary.ndim - 1`` to ``+ary.ndim`` (inclusive). Non-negative
        indices indicate locations before the corresponding axis:
        ``pos=0`` means to insert at the very beginning. Negative
        indices indicate locations after the corresponding axis:
        ``pos=-1`` means to insert at the very end. 0 and -1 are always
        guaranteed to work. Any other number will depend on the
        dimensions of the existing array. Default is 0.

    Returns
    -------
    res : ndarray
        An array with ``res.ndim >= ndim``. A view is returned for array
        inputs. Dimensions are prepended if `pos` is 0, so for example,
        a 1-D array of shape ``(N,)`` with ``ndim=4`` becomes a view of
        shape ``(1, 1, 1, N)``. Dimensions are appended if `pos` is -1,
        so for example a 2-D array of shape ``(M, N)`` becomes a view of
        shape ``(M, N, 1, 1)`` when ``ndim=4``.

    See Also
    --------
    np.atleast_1d, np.atleast_2d, np.atleast_3d

    Notes
    -----
    Taken from https://github.com/numpy/numpy/pull/7804.

    This function does not follow the convention of the other atleast_*d
    functions in numpy in that it only accepts a single array argument.
    To process multiple arrays, use a comprehension or loop around the
    function call. See examples below.
    Setting ``pos=0`` is equivalent to how the array would be
    interpreted by numpy's broadcasting rules. There is no need to call
    this function for simple broadcasting. This is also roughly
    (but not exactly) equivalent to
    ``np.array(ary, copy=False, subok=True, ndmin=ndim)``.
    It is easy to create functions for specific dimensions similar to
    the other atleast_*d functions using Python's `functools.partial`
    function. An example is shown below.
    """
    ary = np.array(ary, copy=False, subok=True)
    if ary.ndim:
        pos = np.normalize_axis_index(pos, ary.ndim + 1)
    extra = ndim - ary.ndim
    if extra > 0:
        ind = pos * (slice(None),) + extra * (None,) + (Ellipsis,)
        ary = ary[ind]
    return ary


def ensure_number_array(arr, shape: Tuple = None, ndim: int = None, size=None,
                        accept_sparse=True) -> Union[np.ndarray, spmatrix]:
    return ensure_array(arr, shape=shape, ndim=ndim, size=size, dtype=np.number, accept_sparse=accept_sparse)


def ensure_integer_array(arr, shape: Tuple = None, ndim: int = None, size=None,
                         accept_sparse=True) -> Union[np.ndarray, spmatrix]:
    return ensure_array(arr, shape=shape, ndim=ndim, size=size, dtype=np.integer, accept_sparse=accept_sparse)


def ensure_floating_array(arr, shape: Tuple = None, ndim: int = None, size=None,
                          accept_sparse=True) -> Union[np.ndarray, spmatrix]:
    return ensure_array(arr, shape=shape, ndim=ndim, size=size, dtype=np.floating, accept_sparse=accept_sparse)


def ensure_array(arr, shape: Optional[Tuple] = None, ndim: Optional[int] = None,
                 dtype=None, size=None, accept_sparse=True) -> Union[np.ndarray, spmatrix]:
    if issparse(arr) and not accept_sparse:
        arr = arr.toarray()
    else:
        if isinstance(arr, set):
            if dtype is not None:
                arr = np.fromiter(arr, dtype, len(arr))
            else:
                arr = np.asarray(list(arr))
        if not isinstance(arr, np.ndarray) and not issparse(arr):
            arr = np.asanyarray(arr)
            if ndim is not None and arr.ndim < ndim:
                arr = atleast_nd(arr, ndim=ndim)

    if shape is not None and arr.shape != shape:
        raise ValueError(f"Shape of provided array was {arr.shape} != {shape}")
    if ndim is not None and arr.ndim != ndim:
        raise ValueError(f"ndim of provided array was {arr.ndim} != {ndim}")
    if size is not None and np.size(arr) != size:
        raise ValueError(f"size of provided array was {np.size(arr)} != {size}")
    if dtype is not None and not np.issubdtype(arr.dtype, dtype):
        raise ValueError(f"Array got incompatible dtype: {arr.dtype} is not a subtype of {dtype}.")
    return arr


def ensure_traj_list(trajs, dtype=None):
    """Makes sure that trajs is a list of trajectories (array of dtype)"""
    if len(trajs) > 0 and not isinstance(trajs, (list, tuple)) and (dtype is None or not isinstance(trajs[0], dtype)):
        return [ensure_floating_array(trajs) if dtype is None else ensure_array(trajs, dtype=dtype)]
    return [ensure_floating_array(t) if dtype is None else ensure_array(t, dtype=dtype) for t in trajs]


def ensure_dtraj_list(dtrajs):
    """Makes sure that dtrajs is a list of discrete trajectories (array of int)"""
    if len(dtrajs) > 0 and isinstance(dtrajs[0], numbers.Real):
        return [ensure_integer_array(dtrajs)]
    return [ensure_integer_array(t) for t in dtrajs]


def ensure_timeseries_data(input_data) -> List[np.ndarray]:
    r""" Ensures that the input data is a time series. This means it must be an iterable of ndarrays or an ndarray
    with a consistent dtype.

    Parameters
    ----------
    input_data : ndarray or list of ndarray
        the input data

    Returns
    -------
    data : list of np.ndarray
        timeseries data
    """
    if isinstance(input_data, np.ndarray) and input_data.ndim >= 3:
        input_data = [x for x in input_data]
    if not isinstance(input_data, (list, tuple)):
        input_data = [input_data]
    grouped = itertools.groupby(input_data, type)
    unique_types = [t for t, _ in grouped]
    if len(unique_types) > 1:
        raise ValueError("All arrays must be of same dtype, but got dtypes {}".format(unique_types))
    assert isinstance(input_data, (list, tuple))
    return input_data


def is_timelagged_dataset(dataset) -> bool:
    r""" Checks whether an object can be ducktyped into a (time-lagged) dataset.

    In particular, this method checks for presence of attributes

    * `__len__`
    * `__getitem__`

    and, if length is positive, checks whether the first element is a tuple of length 2

    Parameters
    ----------
    dataset

    Returns
    -------
    is_dataset : bool
        Whether the input is a time-lagged dataset.
    """
    has_len_and_getitem = hasattr(dataset, '__len__') and hasattr(dataset, '__getitem__')
    if has_len_and_getitem:
        if len(dataset) > 0:
            first_elem = dataset[0]
            return isinstance(first_elem, tuple) and len(first_elem) == 2
        else:
            return False
    return False


def to_dataset(data: Union[TimeLaggedDataset, Tuple[np.ndarray, np.ndarray], np.ndarray],
               lagtime: Optional[int] = None):
    r"""Converts input data to a TimeLaggedDataset if possible, otherwise assumes that data implements `__len__` as well
    as `__getitem__`, where `__getitem__` yields a tuple of data.

    The possible cases are:

    * input data is already a time-lagged dataset, then return immediately (see :meth:`is_timelagged_dataset`).
    * input data is a tuple of (X, Y), where X and Y are ndarrays - in this case they are interpreted as time-lagged
      versions of another, i.e., :math:`Y_i = \mathcal{g}(X_i)`, where :math:`g(\cdot )` describes the temporal
      evolution. In this case lagtime is ignored.
    * input data is a list of trajectories, in this case a concatenated :class:`TrajectoriesDataset` is created
    * input is a ndarray, in this case `Y[i] = X[i+lagtime]` and the result is a dataset of length `len(data) - lagtime`

    Parameters
    ----------
    data : TimeLaggedDataset or tuple of arrays or array
        Input data.
    lagtime : int, optional, default=None
        Lagtime, only is considered if input is array.

    Returns
    -------
    dataset
        A dataset based on input arguments.

    Raises
    ------
    ValueError
        If data is single array but no lagtime is provided or input is list or tuple of length not equal to 2

    Examples
    --------
    Create dataset via trajectory + lagtime

    >>> data = np.arange(0, 6)
    >>> dataset = to_dataset(data, lagtime=1)
    >>> print(dataset[:])
    (array([0, 1, 2, 3, 4]), array([1, 2, 3, 4, 5]))

    Create dataset via corresponding data matrices

    >>> data_instantaneous = np.array([0, 1, 2, 3])
    >>> data_timelagged = np.zeros((4, 2))
    >>> dataset = to_dataset((data_instantaneous, data_timelagged))

    Printing instantaneous data

    >>> print(dataset[:][0])
    [0 1 2 3]

    Printing timelagged data

    >>> print(dataset[:][1])
    [[0. 0.]
     [0. 0.]
     [0. 0.]
     [0. 0.]]
    """
    if isinstance(data, np.ndarray) and data.ndim >= 3:
        data = [x for x in data]
    if isinstance(data, tuple):
        if len(data) != 2:
            raise ValueError(f"If data is provided as tuple the length must be 2 but was {len(data)}.")
        return TimeLaggedDataset(*data)
    if isinstance(data, np.ndarray):
        if lagtime is None:
            raise ValueError("In case data is a single trajectory the lagtime must be given.")
        return TrajectoryDataset(lagtime, data)
    if isinstance(data, list) and len(data) > 0 and isinstance(data[0], np.ndarray):
        data = ensure_timeseries_data(data)
        return TrajectoriesDataset.from_numpy(lagtime, data)
    assert hasattr(data, '__len__') and len(data) > 0, "Data is empty."
    assert is_timelagged_dataset(data), \
        "Data is not a time-lagged dataset, i.e., yielding tuples of instantaneous and time-lagged data. " \
        "In case of multiple trajectories, deeptime.util.data.TrajectoriesDataset may be used."
    return data



import numpy as np


def variable_cols(data: np.ndarray, tol=0.0, min_constant=0):
    """ Evaluates which columns are constant (0) or variable (1)

    Parameters
    ----------
    data : ndarray
        Matrix whose columns will be checked for constant or variable.
    tol : float
        Tolerance for float-matrices. When set to 0 only equal columns with
        values will be considered constant. When set to a positive value,
        columns where all elements have absolute differences to the first
        element of that column are considered constant.
    min_constant : int
        Minimal number of constant columns to resume operation. If at one
        point the number of constant columns drops below min_constant, the
        computation will stop and all columns will be assumed to be variable.
        In this case, an all-True array will be returned.

    Returns
    -------
    variable : bool-array
        Array with number of elements equal to the columns. True: column is
        variable / non-constant. False: column is constant.

    """
    from ._covartools import variable_cols as impl
    # prepare column array
    cols = np.zeros(data.shape[1], dtype=bool, order='C')
    completed = impl(cols, data, tol, min_constant)

    # if interrupted, return all ones. Otherwise return the variable columns as bool array
    return cols if completed == 1 else np.ones_like(cols, dtype=bool)



from scipy.sparse import coo_matrix


def remove_negative_entries(input_matrix):
    r"""Remove all negative entries from sparse matrix.

    Parameters
    ----------
    input_matrix : (M, M) scipy.sparse matrix
        Input matrix

    Returns
    -------
    non_negative_mat : (M, M) scipy.sparse matrix
        Input matrix with negative entries set to zero.
    """
    input_matrix = input_matrix.tocoo()

    data = input_matrix.data
    row = input_matrix.row
    col = input_matrix.col

    pos = data > 0.0
    datap = data[pos]
    rowp = row[pos]
    colp = col[pos]
    return coo_matrix((datap, (rowp, colp)), shape=input_matrix.shape)



r"""
.. currentmodule: deeptime.util

===============================================================================
Data utilities
===============================================================================

.. autosummary::
    :toctree: generated/
    :template: class_nomodule.rst

    data.timeshifted_split
    data.TimeLaggedDataset
    data.TimeLaggedConcatDataset
    data.TrajectoryDataset
    data.TrajectoriesDataset
    data.ConcatDataset

===============================================================================
Statistics utilities
===============================================================================

.. autosummary::
    :toctree: generated/
    :template: class_nomodule.rst

    QuantityStatistics
    confidence_interval
    LaggedModelValidator

===============================================================================
Type utilities
===============================================================================

.. autosummary::
    :toctree: generated/
    :template: class_nomodule.rst

    types.to_dataset
    types.is_timelagged_dataset
    types.atleast_nd


===============================================================================
Other utilities
===============================================================================
.. autosummary::
    :toctree: generated/
    :template: class_nomodule.rst

    parallel.handle_n_jobs
    decorators.cached_property
    decorators.plotting_function
"""

from .stats import QuantityStatistics, confidence_interval
from ._validation import LaggedModelValidator

from . import data
from . import types



import warnings

import numpy as np

from deeptime.util.exceptions import NotConvergedWarning


def mle_trev(C, maxerr=1.0e-12, maxiter=int(1.0E6), warn_not_converged=True, return_statdist=False,
             eps_mu=1.0e-15):
    from ._mle_bindings import mle_trev_dense
    from ...analysis import is_connected

    assert maxerr > 0, 'maxerr must be positive'
    assert maxiter > 0, 'maxiter must be positive'
    assert C.shape[0] == C.shape[1], 'C must be a square matrix.'
    assert is_connected(C, directed=True), 'C must be strongly connected'

    if C.dtype not in (np.float32, np.float64, np.longdouble):
        C = C.astype(np.float64)
    dtype = C.dtype

    C_sum = C.sum(axis=1).astype(dtype, order='C', copy=False)
    CCt = (C + C.T).astype(dtype, order='C', copy=False)
    T = np.zeros(C.shape, dtype=dtype, order='C')
    mu = np.zeros(C.shape[0], dtype=dtype, order='C')

    code = mle_trev_dense(T, CCt, C_sum, CCt.shape[0], maxerr, maxiter, mu, eps_mu)
    if code == -5 and warn_not_converged:
        warnings.warn('Reversible transition matrix estimation didn\'t converge.',
                      NotConvergedWarning)

    if return_statdist:
        return T, mu
    else:
        return T


def mle_trev_given_pi(C, mu, maxerr=1.0E-12, maxiter=1000000):
    from ._mle_bindings import mle_trev_given_pi_dense
    from ...analysis import is_connected

    assert maxerr > 0, 'maxerr must be positive'
    assert maxiter > 0, 'maxiter must be positive'
    assert is_connected(C, directed=False), 'C must be (weakly) connected'

    dtype = C.dtype
    if dtype not in (np.float32, np.float64, np.longdouble):
        dtype = np.float64

    c_C = C.astype(dtype, order='C', copy=False)
    c_mu = mu.astype(dtype, order='C', copy=False)

    assert c_C.shape[0] == c_C.shape[1] == c_mu.shape[0], 'Dimensions of C and mu don\'t agree.'

    T = np.zeros_like(c_C, dtype=dtype, order='C')

    code = mle_trev_given_pi_dense(T, c_C, c_mu, C.shape[0], maxerr, maxiter)

    if code == -5:
        warnings.warn('Reversible transition matrix estimation with fixed stationary distribution didn\'t converge.',
                      NotConvergedWarning)
    return T



import numpy as np


def from_data(trajs, n_hidden_states, reversible):
    r""" Makes an initial guess :class:`HMM <HiddenMarkovModel>` with Gaussian output model.

    To this end, a Gaussian mixture model is estimated using `scikit-learn <https://scikit-learn.org/>`_.

    Parameters
    ----------
    trajs : array_like or list of array_like
        Trajectories which are used for making the initial guess.
    n_hidden_states : int
        Number of hidden states.
    reversible : bool
        Whether the hidden transition matrix is estimated so that it is reversible.

    Returns
    -------
    hmm_init : HiddenMarkovModel
        An initial guess for the HMM

    See Also
    --------
    deeptime.markov.hmm.GaussianOutputModel : The type of output model this heuristic uses.
    deeptime.markov.hmm.init.discrete.metastable_from_data
    deeptime.markov.hmm.init.discrete.metastable_from_msm
    """
    from deeptime.markov.hmm import HiddenMarkovModel, GaussianOutputModel
    from sklearn.mixture import GaussianMixture
    import deeptime.markov.tools.estimation as msmest
    import deeptime.markov.tools.analysis as msmana
    from deeptime.util.types import ensure_timeseries_data

    trajs = ensure_timeseries_data(trajs)
    collected_observations = np.concatenate(trajs)
    if collected_observations.ndim == 1:
        collected_observations = collected_observations[..., None]
    gmm = GaussianMixture(n_components=n_hidden_states)
    gmm.fit(collected_observations)
    output_model = GaussianOutputModel(n_hidden_states, means=gmm.means_[:, 0], sigmas=np.sqrt(gmm.covariances_[:, 0]))

    # Compute fractional state memberships.
    Nij = np.zeros((n_hidden_states, n_hidden_states))
    for o_t in trajs:
        # length of trajectory
        T = o_t.shape[0]
        # output probability
        pobs = output_model.to_state_probability_trajectory(o_t)
        # normalize
        pobs /= pobs.sum(axis=1)[:, None]
        # Accumulate fractional transition counts from this trajectory.
        for t in range(T - 1):
            Nij += np.outer(pobs[t, :], pobs[t + 1, :])

    # Compute transition matrix maximum likelihood estimate.
    transition_matrix = msmest.transition_matrix(Nij, reversible=reversible)
    initial_distribution = msmana.stationary_distribution(transition_matrix)
    return HiddenMarkovModel(transition_model=transition_matrix, output_model=output_model,
                             initial_distribution=initial_distribution)



r"""
.. currentmodule: deeptime.kernels

.. autosummary::
    :toctree: generated/
    :template: class_nomodule.rst

    Kernel
    GaussianKernel
    GeneralizedGaussianKernel
    LaplacianKernel
    PolynomialKernel


.. rubric:: PyTorch-ready kernels

.. autosummary::
    :toctree: generated/
    :template: class_nomodule.rst

    TorchGaussianKernel

.. rubric:: Utils

.. autosummary::
    :toctree: generated/
    :template: class_nomodule.rst

    is_torch_kernel
"""

from ._base import Kernel, is_torch_kernel
from ._kernels import GaussianKernel, GeneralizedGaussianKernel, LaplacianKernel, PolynomialKernel

from ..util.platform import module_available
if module_available("torch"):
    from ._kernels_torch import TorchGaussianKernel
del module_available



import warnings
from typing import List

import numpy as np

from ..base import Model


def pcca(P, m, stationary_distribution=None):
    """PCCA+ spectral clustering method with optimized memberships.

    Implementation according to :footcite:`roblitz2013fuzzy`.
    Clusters the first m eigenvectors of a transition matrix in order to cluster the states.
    This function does not assume that the transition matrix is fully connected. Disconnected sets
    will automatically define the first metastable states, with perfect membership assignments.

    Parameters
    ----------
    P : ndarray (n,n)
        Transition matrix.
    m : int
        Number of clusters to group to.
    stationary_distribution : ndarray(n,), optional, default=None
        Stationary distribution over the full state space, can be given if already computed.

    References
    ----------
    .. footbibliography::
    """
    if m <= 0 or m > P.shape[0]:
        raise ValueError("Number of metastable sets must be larger than 0 and can be at most as large as the number "
                         "of states.")
    assert 0 < m <= P.shape[0]
    from scipy.sparse import issparse
    if issparse(P):
        warnings.warn('PCCA is only implemented for dense matrices, '
                      'converting sparse transition matrix to dense ndarray.', stacklevel=2)
        P = P.toarray()

    # stationary distribution
    if stationary_distribution is None:
        from deeptime.markov.tools.analysis import stationary_distribution
        pi = stationary_distribution(P)
    else:
        pi = stationary_distribution

    # memberships
    from .tools.analysis.dense._pcca import pcca as _algorithm_impl
    M = _algorithm_impl(P, m, pi)

    # coarse-grained stationary distribution
    pi_coarse = np.dot(M.T, pi)

    # HMM output matrix
    B = np.linalg.multi_dot([np.diag(1.0 / pi_coarse), M.T, np.diag(pi)])
    # renormalize B to make it row-stochastic
    B /= B.sum(axis=1)[:, None]

    # coarse-grained transition matrix
    W = np.linalg.inv(np.dot(M.T, M))
    A = np.dot(np.dot(M.T, P), M)
    P_coarse = np.dot(W, A)

    # symmetrize and renormalize to eliminate numerical errors
    X = np.dot(np.diag(pi_coarse), P_coarse)
    # and normalize
    P_coarse = X / X.sum(axis=1)[:, None]

    return PCCAModel(P_coarse, pi_coarse, M, B)


class PCCAModel(Model):
    """
    Model for PCCA+ spectral clustering method with optimized memberships.

    PCCA+ spectral clustering is described in :footcite:`roblitz2013fuzzy`.
    Clusters the first m eigenvectors of a transition matrix in order to cluster the states.
    This function does not assume that the transition matrix is fully connected. Disconnected sets
    will automatically define the first metastable states, with perfect membership assignments.

    Parameters
    ----------
    transition_matrix_coarse : ndarray (n,n)
        Coarse transition matrix.
    pi_coarse : ndarray (n,)
        Coarse stationary distribution
    memberships : ndarray (n,m)
        The pcca memberships to clusters
    metastable_distributions : ndarray (m, n)
        metastable distributions

    See Also
    --------
    pcca : Method that produces this type of model.
    msm.MarkovStateModel.pcca : Coarse-grain with PCCA+ from already existing MSM instance.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, transition_matrix_coarse: np.ndarray, pi_coarse: np.ndarray, memberships: np.ndarray,
                 metastable_distributions: np.ndarray):
        super().__init__()
        self._transition_matrix_coarse = transition_matrix_coarse
        self._pi_coarse = pi_coarse
        self._memberships = memberships
        self._metastable_distributions = metastable_distributions
        self._m = self._memberships.shape[1]

    @property
    def n_metastable(self):
        r""" Number of metastable states. """
        return self._m

    @property
    def memberships(self):
        r""" Probabilities of MarkovStateModel states to belong to a metastable state by PCCA+

        Returns the memberships of active set states to metastable sets.

        Returns
        -------
        M : ndarray((n,m))
            A matrix containing the probability or membership of each state to be
            assigned to each metastable set, i.e. p(metastable | state).
            The row sums of M are 1.
        """
        return self._memberships

    @property
    def metastable_distributions(self):
        r""" Probability of metastable states to visit an MarkovStateModel state by PCCA+

        Returns the probability distributions of active set states within
        each metastable set by combining the PCCA+ method with
        Bayesian inversion as described in :footcite:`noe2013projected`.

        Returns
        -------
        p_out : ndarray (m,n)
            A matrix containing the probability distribution of each active set
            state, given that we are in one of the m metastable sets,
            i.e. p(state | metastable). The row sums of p_out are 1.
        """
        return self._metastable_distributions

    @property
    def coarse_grained_transition_matrix(self):
        r""" Coarse grained transition matrix with :attr:`n_metastable` states. """
        return self._transition_matrix_coarse

    @property
    def coarse_grained_stationary_probability(self):
        r""" Stationary distribution for :attr:`coarse_grained_transition_matrix`. """
        return self._pi_coarse

    @property
    def assignments(self) -> np.ndarray:
        """ Assignment of states to metastable sets using PCCA++

        Computes the assignment to metastable sets for active set states using
        the PCCA++ method :footcite:`roblitz2013fuzzy`.

        This is only recommended for visualization purposes. You *cannot* compute
        any actual quantity of the coarse-grained kinetics without employing the
        fuzzy memberships!

        Returns
        -------
        assignments : ndarray (n,)
            For each MarkovStateModel state, the metastable state it is located in.

        """
        return np.argmax(self.memberships, axis=1)

    @property
    def sets(self) -> List[np.ndarray]:
        """ Metastable sets using PCCA+

        Computes the metastable sets of active set states within each
        metastable set using the PCCA+ method :footcite:`roblitz2013fuzzy`.

        This is only recommended for visualization purposes. You *cannot*
        compute any actual quantity of the coarse-grained kinetics without
        employing the fuzzy memberships!

        Returns
        -------
        sets : list of ndarray
            A list of length equal to metastable states. Each element is an
            array with microstate indexes contained in it
        """
        res = []
        assignment = self.assignments
        for i in range(self.n_metastable):
            res.append(np.where(assignment == i)[0])
        return res



import itertools
from typing import Optional

import numpy as np
from scipy.linalg import eig

from ..base import Estimator, Model, Transformer, EstimatorTransformer
from ..basis import Observable
from ..util.data import timeshifted_split
from ..numeric import spd_inv_split, sort_eigs, spd_inv_sqrt
from .util import running_covar
from ..util.types import ensure_timeseries_data

__all__ = ['Covariance', 'CovarianceModel', 'KoopmanWeightingEstimator', 'KoopmanWeightingModel']

__author__ = 'paul, nueske, marscher, clonker'


class WhiteningTransform(Observable):
    r""" Transformation of data into a whitened space. It is assumed that for a covariance matrix :math:`C` the
    square-root inverse :math:`C^{-1/2}` was already computed. Optionally a mean :math:`\mu` can be provided.
    This yields the transformation

    .. math::
        y = C^{-1/2}(x-\mu).

    Parameters
    ----------
    sqrt_inv_cov : (n, k) ndarray
        Square-root inverse of covariance matrix.
    mean : (n, ) ndarray, optional, default=None
        The mean if it should be subtracted.
    dim : int, optional, default=None
        Additional restriction in the dimension, removes all but the first `dim` components of the output.

    See Also
    --------
    deeptime.numeric.spd_inv_sqrt : Method to obtain (regularized) inverses of covariance matrices.
    """

    def __init__(self, sqrt_inv_cov: np.ndarray, mean: Optional[np.ndarray] = None, dim: Optional[int] = None):
        self.sqrt_inv_cov = sqrt_inv_cov
        self.mean = mean
        self.dim = dim

    def _evaluate(self, x: np.ndarray):
        if self.mean is not None:
            x = x - self.mean
        return x @ self.sqrt_inv_cov[..., :self.dim]


class CovarianceModel(Model):
    r""" A model which in particular carries the estimated covariances, means from a :class:`Covariance`.

    Parameters
    ----------
    cov_00 : (n, n) ndarray, optional, default=None
        The instantaneous covariances if computed (see :attr:`Covariance.compute_c00`).
    cov_0t : (n, n) ndarray, optional, default=None
        The time-lagged covariances if computed (see :attr:`Covariance.compute_c0t`).
    cov_tt : (n, n) ndarray, optional, default=None
        The time-lagged instantaneous covariances if computed (see :attr:`Covariance.compute_ctt`).
    mean_0 : (n,) ndarray, optional, default=None
        The instantaneous means if computed.
    mean_t : (n,) ndarray, optional, default=None
        The time-shifted means if computed.
    bessels_correction : bool, optional, default=True
        Whether Bessel's correction was used during estimation.
    lagtime : int, default=None
        The lagtime that was used during estimation.
    data_mean_removed : bool, default=False
        Whether the data mean was removed. This can have an influence on the effective VAMP score.
    """

    def __init__(self, cov_00: Optional[np.ndarray] = None, cov_0t: Optional[np.ndarray] = None,
                 cov_tt: Optional[np.ndarray] = None, mean_0: Optional[np.ndarray] = None,
                 mean_t: Optional[np.ndarray] = None, bessels_correction: bool = True,
                 symmetrized: bool = False, lagtime: Optional[int] = None, data_mean_removed: bool = False):
        super(CovarianceModel, self).__init__()
        self._cov_00 = cov_00
        self._cov_0t = cov_0t
        self._cov_tt = cov_tt
        self._mean_0 = mean_0
        self._mean_t = mean_t
        self._bessel = bessels_correction
        self._lagtime = lagtime
        self._symmetrized = symmetrized
        self._data_mean_removed = data_mean_removed

    @property
    def cov_00(self) -> Optional[np.ndarray]:
        r""" The instantaneous covariances.

        :type: (n, n) ndarray or None
        """
        return self._cov_00

    @property
    def cov_0t(self) -> Optional[np.ndarray]:
        r""" The time-shifted covariances.

        :type: (n, n) ndarray or None
        """
        return self._cov_0t

    @property
    def cov_tt(self) -> Optional[np.ndarray]:
        r""" The time-shifted instantaneous covariances.

        :type: (n, n) ndarray or None
        """
        return self._cov_tt

    @property
    def mean_0(self) -> Optional[np.ndarray]:
        r""" The instantaneous means.

        :type: (n,) ndarray or None
        """
        return self._mean_0

    @property
    def mean_t(self) -> Optional[np.ndarray]:
        r""" The time-shifted means.

        :type: (n,) ndarray or None
        """
        return self._mean_t

    @property
    def bessels_correction(self) -> bool:
        r""" Whether Bessel's correction was applied during estimation.

        :type: bool
        """
        return self._bessel

    @property
    def lagtime(self) -> Optional[int]:
        r""" The lagtime at which estimation was performed.

        :type: int or None
        """
        return self._lagtime

    @property
    def symmetrized(self) -> bool:
        r""" Whether correlations and second moments are symmetrized in time.

        :type: bool
        """
        return self._symmetrized

    def whiten(self, data: np.ndarray, epsilon=1e-10, method='QR') -> np.ndarray:
        r"""Whiten a (T, N)-shaped chunk of data by transforming it into the PCA basis. In case of rank deficiency
        this reduces the dimension.

        Parameters
        ----------
        data : (T,N) ndarray
            The data to be whitened.
        epsilon : float, optional, default=1e-10
            Truncation parameter. See :meth:`deeptime.numeric.spd_inv_sqrt`.
        method : str, optional, default='QR'
            Decomposition method. See :meth:`deeptime.numeric.spd_inv_sqrt`.

        Returns
        -------
        whitened_data : (T, n) ndarray
            Whitened data.
        """
        assert self.cov_00 is not None and self.mean_0 is not None
        projection = np.atleast_2d(spd_inv_sqrt(self.cov_00, epsilon=epsilon, method=method))
        whitened_data = (data - self.mean_0[None, ...]) @ projection.T
        return whitened_data

    @property
    def data_mean_removed(self) -> bool:
        r"""Whether the data mean was removed.

        :type: bool
        """
        return self._data_mean_removed


class Covariance(Estimator):
    r"""Compute (potentially lagged) covariances between data in an online fashion.

    This means computing

    .. math:: \mathrm{cov}[ X_t, Y_t ] = \mathbb{E}[(X_t - \mathbb{E}[X_t])(Y_t - \mathbb{E}[Y_t])],

    where :math:`X_t` and :math:`Y_t` are contiguous blocks of frames from the timeseries data. The estimator
    implements the online algorithm proposed in :footcite:`chan1982updating`, report available
    `here <http://i.stanford.edu/pub/cstr/reports/cs/tr/79/773/CS-TR-79-773.pdf>`__.

    Parameters
    ----------
    lagtime : int, default=0
        The lagtime, must be :math:`\geq 0` .
    compute_c00 : bool, optional, default=True
        Compute instantaneous correlations over the first part of the data. If :attr:`lagtime` ==0,
        use all of the data.
    compute_c0t : bool, optional, default=False
        Compute lagged correlations. Does not work with :attr:`lagtime` ==0.
    compute_ctt : bool, optional, default=False
        Compute instantaneous covariance over the time-shifted chunks of the data.
        Does not work with :attr:`lagtime` ==0.
    remove_data_mean : bool, optional, default=False
        Subtract the sample mean from the time series (mean-free correlations).
    reversible : bool, optional, default=False
        Symmetrize correlations, i.e., use estimates defined by :math:`\sum_t X_t + Y_t` and second
        moment matrices defined by :math:`X_t^\top X_t + Y_t^\top Y_t` and :math:`Y_t^\top X_t + X_t^\top Y_t` .
    bessels_correction : bool, optional, default=True
        Use Bessel's correction for correlations in order to use an unbiased estimator.
    sparse_mode : str, optional, default='auto'
        one of:
            * 'dense' : always use dense mode
            * 'auto' : automatic
            * 'sparse' : always use sparse mode if possible
    ncov : int, optional, default=5
        Depth of moment storage. Moments computed from each chunk will be combined with Moments of similar
        statistical weight using the pairwise combination algorithm
        described in :footcite:`chan1982updating`.
    diag_only: bool
        If True, the computation is restricted to the diagonal entries (autocorrelations) only.
    model : CovarianceModel, optional, default=None
        A model instance with which the estimator can be initialized.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, lagtime: int = 0, compute_c00: bool = True, compute_c0t: bool = False,
                 compute_ctt: bool = False, remove_data_mean: bool = False, reversible: bool = False,
                 bessels_correction: bool = True, sparse_mode: str = 'auto', ncov: int = 5, diag_only: bool = False,
                 model=None):
        super(Covariance, self).__init__(model=model)
        self.lagtime = lagtime
        self.compute_c00 = compute_c00
        self.compute_c0t = compute_c0t
        self.compute_ctt = compute_ctt
        self.remove_data_mean = remove_data_mean
        self.reversible = reversible
        self.bessels_correction = bessels_correction
        self.diag_only = diag_only
        self.ncov = ncov
        self.sparse_mode = sparse_mode
        self._dirty = False  # Flag which gets set to true once fit/partial_fit was called and the model was not fetched

        self._rc = running_covar(xx=self.compute_c00, xy=self.compute_c0t, yy=self.compute_ctt,
                                 remove_mean=self.remove_data_mean, symmetrize=self.reversible,
                                 sparse_mode=self.sparse_mode, modify_data=False, diag_only=self.diag_only,
                                 nsave=ncov)

    @property
    def ncov(self) -> int:
        r""" The depth of the moments storage.

        :type: int
        """
        return self._ncov

    @ncov.setter
    def ncov(self, value: int):
        self._ncov = value

    @property
    def diag_only(self) -> bool:
        r"""" Whether the computation should be restricted to diagonal entries (autocorrelations) only.

        :type: bool
        """
        return self._diag_only

    @diag_only.setter
    def diag_only(self, value: bool):
        self._diag_only = value

    @property
    def sparse_mode(self) -> str:
        r""" The sparse mode of the estimator. Can be one of 'auto', 'sparse', and 'dense'.

        :type: str
        """
        return self._sparse_mode

    @sparse_mode.setter
    def sparse_mode(self, value: str):
        valid_modes = ('auto', 'sparse', 'dense')
        if self.diag_only and value != 'dense':
            if value == 'sparse':
                import warnings
                warnings.warn('Computing diagonal entries only is not implemented for sparse mode. Switching to dense '
                              'mode.')
            value = 'dense'
        if value not in valid_modes:
            raise ValueError("Unknown sparse mode: {}, must be one of {}.".format(value, valid_modes))
        self._sparse_mode = value

    @property
    def bessels_correction(self) -> bool:
        r""" Whether to apply Bessel's correction for an unbiased estimator.

        :type: bool
        """
        return self._bessels_correction

    @bessels_correction.setter
    def bessels_correction(self, value: bool):
        self._bessels_correction = value

    @property
    def reversible(self) -> bool:
        r"""Whether to symmetrize correlations.

        :type: bool
        """
        return self._reversible

    @reversible.setter
    def reversible(self, value: bool):
        if self.compute_ctt and value:
            raise ValueError("Computing covariances reversibly and also computing cov_tt is meaningless, as then "
                             "cov_tt = cov_00. Please set compute_ctt to False prior setting reversible to True "
                             "if symmetrized covariances are desired.")
        self._reversible = value

    @property
    def remove_data_mean(self) -> bool:
        r""" Whether to remove the sample mean, i.e., compute mean-free correlations.

        :type: bool
        """
        return self._remove_data_mean

    @remove_data_mean.setter
    def remove_data_mean(self, value: bool):
        self._remove_data_mean = value

    @property
    def compute_c00(self) -> bool:
        r""" Whether to compute instantaneous correlations.

        :type: bool
        """
        return self._compute_c00

    @compute_c00.setter
    def compute_c00(self, value: bool):
        self._compute_c00 = value

    @property
    def compute_c0t(self) -> bool:
        r"""Whether to compute time lagged correlations with a defined :attr:`lagtime`.

        :type: bool
        """
        return self._compute_c0t

    @compute_c0t.setter
    def compute_c0t(self, value: bool):
        self._compute_c0t = value

    @property
    def compute_ctt(self) -> bool:
        r"""Whether to compute instantaneous correlations over the time-shifted chunks of the data.

        :type: bool
        """
        return self._compute_ctt

    @compute_ctt.setter
    def compute_ctt(self, value: bool):
        self._compute_ctt = value

    @property
    def lagtime(self) -> int:
        r"""
        The lagtime of the estimator. This attribute determines how big the temporal difference for timelagged
        autocorrelations are.

        :getter: Yields the currently selected lagtime.
        :setter: Sets a new lagtime, must be :math:`\geq 0`, for :attr:`compute_c0t` and :attr:`compute_ctt` it
                 must be :math:`> 0`.
        :type: int
        """
        return self._lagtime

    @lagtime.setter
    def lagtime(self, value: Optional[int]):
        if value is not None and value < 0:
            raise ValueError("Negative lagtime are not supported.")
        self._lagtime = value

    @property
    def is_lagged(self) -> bool:
        r""" Determines whether this estimator also computes time-lagged covariances.

        :type: bool
        """
        return self.compute_c0t or self.compute_ctt

    def fit(self, data, lagtime=None, weights=None, n_splits=None, column_selection=None):
        r"""Computes covariances for the input data and produces a new model. If an existing model should be updated,
        call :meth:`partial_fit`.

        Parameters
        ----------
        data : array_like or list of array_like
            The input data. If it is a list of trajectories, all elements of the list must have the same dtype and
            size in the second dimension, i.e., the elements of :code:`[x.shape[1] for x in data]` must all be equal.
        lagtime : int, optional, default=None
            Override for :attr:`lagtime`.
        weights : array_like or list of array_like or object, optional, default=None

            * Optional weights for the input data. Must be of matching shape.

            * Can also be another arbitrary object. The only requirement is that weights possesses a method weights(X),
              that accepts a trajectory X (np.ndarray(T, n)) and returns a vector of
              re-weighting factors (np.ndarray(T,)). See, e.g.,

              * :class:`KoopmanEstimator <deeptime.covariance.KoopmanEstimator>`

        n_splits : int, optional, default=None
            The number of times the data is split uniformly when performing the covariance estimation. If no value
            is given, it estimates the number of splits by :code:`min(trajectory_lengths) // 100` if the shortest
            trajectory contains at least 1000 frames. Otherwise, the number of splits is set to one.
        column_selection : ndarray, optional, default=None
            Columns of the trajectories to restrict estimation to. Must be given in terms of an index array.

        Returns
        -------
        self : Covariance
            Reference to self.
        """
        data = ensure_timeseries_data(data)

        self._model = None
        self._rc.clear()

        n_splits = 1 if n_splits is None else n_splits

        if lagtime is None:
            lagtime = self.lagtime
        else:
            self.lagtime = lagtime
        assert lagtime is not None

        lazy_weights = False
        wsplit = itertools.repeat((None, None))  # tuple of None that can be unpacked into two Nones

        if weights is not None:
            if hasattr(weights, 'weights'):
                lazy_weights = True
            elif len(np.atleast_1d(weights)) != len(data[0]):
                raise ValueError(
                    "Weights have incompatible shape "
                    f"(#weights={len(weights) if weights is not None else None} != {len(data[0])}=#frames.")
            elif isinstance(weights, np.ndarray):
                wsplit = timeshifted_split(weights, lagtime=lagtime, n_splits=n_splits)

        if self.is_lagged:
            t = 0
            for (x, y), (w, _) in zip(timeshifted_split(data, lagtime=lagtime, n_splits=n_splits), wsplit):
                if lazy_weights:
                    w = weights.weights(x)
                # weights can be longer than actual data
                if isinstance(w, np.ndarray):
                    w = w[:len(x)]
                self.partial_fit((x, y), weights=w, column_selection=column_selection)
                t += len(x)
        else:
            for x in data:
                self.partial_fit(x, weights=weights, column_selection=column_selection)

        return self

    def partial_fit(self, data, weights=None, column_selection=None):
        """ Incrementally update the estimates. For a detailed description of the parameters, see :meth:`fit` with
        the exception of the :code:`data` argument, it must be a ndarray and cannot be a list of ndarray."""
        self._dirty = True
        if self.is_lagged:
            x, y = data
        else:
            x, y = data, None
        if weights is not None and hasattr(weights, 'weights'):
            weights = weights(x)
        try:
            self._rc.add(x, y, column_selection=column_selection, weights=weights)
        except MemoryError:
            raise MemoryError(f'Covariance matrix does not fit into memory. '
                              f'Input is too high-dimensional ({x.shape[1]} dimensions).')
        return self

    def fetch_model(self) -> CovarianceModel:
        r""" Finalizes the covariance computation by aggregating all moment storages.

        Returns
        -------
        model : CovarianceModel
            The covariance model.
        """
        if self._dirty:
            cov_00 = cov_tt = cov_0t = mean_0 = mean_t = None
            if self.compute_c0t:
                cov_0t = self._rc.cov_XY(self.bessels_correction)
            if self.compute_ctt:
                cov_tt = self._rc.cov_YY(self.bessels_correction)
            if self.compute_c00:
                cov_00 = self._rc.cov_XX(self.bessels_correction)

            if self.compute_c00 or self.compute_c0t:
                mean_0 = self._rc.mean_X()
            if self.compute_ctt or self.compute_c0t:
                mean_t = self._rc.mean_Y()
            if self.reversible:
                cov_tt = cov_00
            self._model = CovarianceModel(cov_00=cov_00, cov_0t=cov_0t, cov_tt=cov_tt, mean_0=mean_0, mean_t=mean_t,
                                          bessels_correction=self.bessels_correction, lagtime=self.lagtime,
                                          symmetrized=self.reversible, data_mean_removed=self.remove_data_mean)
            self._dirty = False  # catches multiple calls to fetch_model even though it hasn't changed
        return self._model


class KoopmanWeightingModel(Model, Transformer):
    r""" A model which contains the Koopman operator in a modified basis `(PC|1)` and can transform data into Koopman
    weights.

    Weights are computed according to :footcite:`wu2017variational`.

    Parameters
    ----------
    u : ndarray
        Reweighting vector in input basis
    u_const : float
        Constant offset for reweighting in input basis.
    koopman_operator : ndarray
        Koopman operator in modified basis.
    whitening_transformation : ndarray, optional, default=None
        Whitening transformation.
    covariances : CovarianceModel, optional, default=None
        Estimated covariances.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, u, u_const, koopman_operator, whitening_transformation=None, covariances=None):
        super().__init__()
        self._u = u
        self._u_const = u_const
        self._koopman_operator = koopman_operator
        self._whitening_transformation = whitening_transformation
        self._covariances = covariances

    def weights(self, X):
        r""" Applies reweighting vectors to data, yielding corresponding weights.

        Parameters
        ----------
        X : (T, d) ndarray
            The input data.

        Returns
        -------
        weights : (T, 1) ndarray
            Weights for input data.
        """
        return X.dot(self.weights_input) + self.const_weight_input

    def transform(self, data, **kw):
        r""" Same as :meth:`weights`. """
        return self.weights(data)

    @property
    def weights_input(self) -> np.ndarray:
        r""" Yields the reweighting vector in input basis.

        :type: (T, d) ndarray
        """
        return self._u

    @property
    def const_weight_input(self) -> float:
        r""" Yields the constant offset for reweighting in input basis.

        :type: float
        """
        return self._u_const

    @property
    def koopman_operator(self) -> np.ndarray:
        r""" The Koopman operator in modified basis (PC|1).

        :type: ndarray
        """
        return self._koopman_operator

    @property
    def whitening_transformation(self) -> np.ndarray:
        r""" Estimated whitening transformation for data

        :type: ndarray or None
        """
        return self._whitening_transformation

    @property
    def covariances(self) -> CovarianceModel:
        r""" Covariance model which was used to compute the Koopman model.

        :type: CovarianceModel or None
        """
        return self._covariances


class KoopmanWeightingEstimator(EstimatorTransformer):
    r"""Computes Koopman operator and weights that can be plugged into the :class:`Covariance` estimator.
    The weights are determined by the procedure described in :footcite:`wu2017variational`.

    Parameters
    ----------
    lagtime : int
        The lag time at which the operator is estimated.
    epsilon : float, optional, default=1e-6
        Truncation parameter. Eigenvalues with norms smaller than this cutoff will be removed.
    ncov : int or str, optional, default=infinity
        Depth of moment storage. Per default no moments are collapsed while estimating covariances, perform
        aggregation only at the very end after all data has been processed.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, lagtime, epsilon=1e-6, ncov='inf'):
        super(KoopmanWeightingEstimator, self).__init__()
        self.epsilon = epsilon
        if ncov == 'inf':
            ncov = int(2 ** 10000)
        self._cov = Covariance(lagtime=lagtime, compute_c00=True, compute_c0t=True, remove_data_mean=True,
                               reversible=False, bessels_correction=False, ncov=ncov)

    def fit(self, data, lagtime=None, **kw):
        r""" Fits a new model.

        Parameters
        ----------
        data : (T, d) ndarray
            The input data.
        lagtime : int, optional, default=None
            Optional override for estimator's :attr:`lagtime`.
        **kw
            Ignored keyword args for scikit-learn compatibility.

        Returns
        -------
        self : KoopmanWeightingEstimator
            Reference to self.
        """
        self._model = None
        self._cov.fit(data, lagtime=lagtime)
        return self

    def partial_fit(self, data):
        r""" Updates the current model using a chunk of data.

        Parameters
        ----------
        data : (T, d) ndarray
            A chunk of data.

        Returns
        -------
        self : KoopmanWeightingEstimator
            Reference to self.
        """
        self._cov.partial_fit(data)
        return self

    def transform(self, data, **kw):
        r""" Computes weights for a chunk of data. This requires that a model was :meth:`fit`.

        Parameters
        ----------
        data : (T, d) ndarray
            A chunk of data.
        **kw
            Ignored kwargs.

        Returns
        -------
        weights : (T, 1) ndarray
            Koopman weights.
        """
        return self.fetch_model().transform(data)

    @staticmethod
    def _compute_u(K):
        r"""Estimate an approximation of the ratio of stationary over empirical distribution from the basis.

        Parameters:
        -----------
        K0 : (M+1, M+1) ndarray
            Time-lagged correlation matrix for the whitened and padded data set.

        Returns:
        --------
        weights : (M,) ndarray
            Coefficients of the ratio stationary / empirical distribution from the whitened and expanded basis.
        """
        M = K.shape[0] - 1
        # Compute right and left eigenvectors:
        l, U = eig(K.T)
        l, U = sort_eigs(l, U)
        # Extract the eigenvector for eigenvalue one and normalize:
        u = np.real(U[:, 0])
        v = np.zeros(M + 1)
        v[M] = 1.0
        u = u / np.dot(u, v)
        return u

    def fetch_model(self) -> KoopmanWeightingModel:
        r""" Finalizes the model.

        Returns
        -------
        koopman_model : KoopmanWeightingModel
            The Koopman model, in particular containing operator and weights.
        """
        cov = self._cov.fetch_model()

        R = spd_inv_split(cov.cov_00, epsilon=self.epsilon, canonical_signs=True)
        # Set the new correlation matrix:
        M = R.shape[1]
        K = np.dot(R.T, np.dot(cov.cov_0t, R))
        K = np.vstack((K, np.dot((cov.mean_t - cov.mean_0), R)))
        ex1 = np.zeros((M + 1, 1))
        ex1[M, 0] = 1.0
        K = np.hstack((K, ex1))

        u = self._compute_u(K)
        N = R.shape[0]
        u_input = np.zeros(N + 1)
        u_input[0:N] = R.dot(u[0:-1])  # in input basis
        u_input[N] = u[-1] - cov.mean_0.dot(R.dot(u[0:-1]))

        self._model = KoopmanWeightingModel(u=u_input[:-1], u_const=u_input[-1], koopman_operator=K,
                                            whitening_transformation=R, covariances=cov)

        return self._model

    @property
    def lagtime(self) -> int:
        r""" The lagtime at which the Koopman operator is estimated.

        :getter: Yields the currently configured lagtime.
        :setter: Sets a new lagtime, must be >= 0.
        :type: int
        """
        return self._cov.lagtime

    @lagtime.setter
    def lagtime(self, value: int):
        self._cov.lagtime = value



r"""
============================
Transition path theory tools
============================

.. currentmodule:: deeptime.markov.tools.flux

This module (:mod:`deeptime.markov.tools.flux`) contains functions to compute reactive flux networks and
find dominant reaction pathways in such networks.

Reactive flux
=============

.. autosummary::
   :toctree: generated/

   flux_matrix - TPT flux network
   to_netflux - Netflux from gross flux
   flux_production - Net flux-production for all states
   flux_producers
   flux_consumers
   coarsegrain

Reaction rates and fluxes
=========================

.. autosummary::
   :toctree: generated/

   total_flux
   rate
   mfpt


Pathway decomposition
=====================

.. autosummary::
   :toctree: generated/

   pathways

"""

from .api import flux_matrix, to_netflux, flux_production, flux_producers, flux_consumers, coarsegrain
from .api import total_flux, rate, mfpt
from .api import pathways



r"""
.. currentmodule: deeptime.markov.hmm

The implementations that can be found in this subpackage are based on the `BHMM <https://github.com/bhmm/bhmm>`_
software package.

Output models
-------------

.. autosummary::
    :toctree: generated/
    :template: class_nomodule.rst

    OutputModel
    DiscreteOutputModel
    GaussianOutputModel


Maximum-likelihood estimation of HMMs
-------------------------------------

Since the maximum-likelihood estimation (via Baum-Welch :footcite:`baum1967inequality`) of
:class:`HMMs <deeptime.markov.hmm.HiddenMarkovModel>` is likely to
get stuck in local maxima, a good initial guess is important for a high-quality model. The following methods aim to
provide heuristics which can yield such an initial guess and can be found in :footcite:`noe2013projected`.

For a HMM with a :class:`discrete output model <deeptime.markov.hmm.DiscreteOutputModel>`, the following main
steps are involved:

1. A :class:`markov state model <deeptime.markov.msm.MarkovStateModel>` is estimated from given discrete
   trajectories. This step is optional if the markov model already exists.
2. The estimated MSM transition matrix :math:`P\in\mathbb{R}^{n\times n}` is coarse-grained with PCCA+ into the
   desired number of hidden states :math:`m` and memberships :math:`M\in\mathbb{R}^{n \times m}`. The transition
   matrix is projected into the hidden state space by

   .. math::
        P_\mathrm{coarse} = (M^\top M)^{-1} M^\top P M.

.. autosummary::
    :toctree: generated/
    :template: class_nomodule.rst

    init.discrete.metastable_from_data
    init.discrete.metastable_from_msm

A non data-driven way to initialize a HMM is implemented by setting the transition matrix and initial distribution to
uniform, and drawing a random row-stochastic emission probability matrix:

.. autosummary::
    :toctree: generated/
    :template: class_nomodule.rst

    init.discrete.random_guess

For a HMM with a :class:`gaussian output model <deeptime.markov.hmm.GaussianOutputModel>`, a Gaussian mixture
model is estimated. This particular heuristic requires an installation of `scikit-learn <https://scikit-learn.org/>`_.

.. autosummary::
    :toctree: generated/
    :template: class_nomodule.rst

    init.gaussian.from_data

Estimation and resulting model:

.. autosummary::
    :toctree: generated/
    :template: class_nomodule.rst

    MaximumLikelihoodHMM
    HiddenMarkovModel

Bayesian hidden markov state models
-----------------------------------
Bayesian HMMs can provide confidence estimates. They are estimated by starting from a reference HMM and then use
Gibbs sampling.

See :footcite:`chodera2011bayesian` for a manuscript describing the theory behind using Gibbs
sampling to sample from Bayesian hidden Markov model posteriors.

.. autosummary::
    :toctree: generated/
    :template: class_nomodule.rst

    BayesianHMM
    BayesianHMMPosterior

Utilities
---------

.. autosummary::
    :toctree: generated/
    :template: class_nomodule.rst

    viterbi

References
----------
.. footbibliography::
"""

from ._hidden_markov_model import HiddenMarkovModel, viterbi
from ._maximum_likelihood_hmm import MaximumLikelihoodHMM
from ._bayesian_hmm import BayesianHMM, BayesianHMMPosterior
from ._output_model import OutputModel, DiscreteOutputModel, GaussianOutputModel

from . import init



r"""

=======================
 Analysis API
=======================

"""

import warnings

import numpy as _np
from scipy.sparse import issparse as _issparse
from deeptime.util.types import ensure_number_array, ensure_integer_array, ensure_floating_array

from ._stationary_vector import stationary_distribution
from ._decomposition import timescales_from_eigenvalues

from . import dense
from . import _assessment
from . import _mean_first_passage_time
from . import _decomposition
from . import _fingerprints
from . import _committor
from . import _expectations

__docformat__ = "restructuredtext en"
__authors__ = __author__ = "Benjamin Trendelkamp-Schroer, Martin Scherer, Jan-Hendrik Prinz, Frank Noe"
__copyright__ = "Copyright 2014, Computational Molecular Biology Group, FU-Berlin"
__credits__ = ["Benjamin Trendelkamp-Schroer", "Martin Scherer", "Jan-Hendrik Prinz", "Frank Noe"]

__version__ = "2.0.0"
__maintainer__ = "Martin Scherer"
__email__ = "m.scherer AT fu-berlin DOT de"


_type_not_supported = \
    TypeError("T is not a numpy.ndarray or a scipy.sparse matrix.")


################################################################################
# Assessment tools
################################################################################


def is_transition_matrix(T, tol=1e-12):
    r"""Check if the given matrix is a transition matrix.

    Parameters
    ----------
    T : (M, M) ndarray or scipy.sparse matrix
        Matrix to check
    tol : float (optional)
        Floating point tolerance to check with

    Returns
    -------
    is_transition_matrix : bool
        True, if T is a valid transition matrix, False otherwise

    Notes
    -----
    A valid transition matrix :math:`P=(p_{ij})` has non-negative
    elements, :math:`p_{ij} \geq 0`, and elements of each row sum up
    to one, :math:`\sum_j p_{ij} = 1`. Matrices wit this property are
    also called stochastic matrices.

    Examples
    --------
    >>> import numpy as np
    >>> from deeptime.markov.tools.analysis import is_transition_matrix

    >>> A = np.array([[0.4, 0.5, 0.3], [0.2, 0.4, 0.4], [-1, 1, 1]])
    >>> is_transition_matrix(A)
    False

    >>> T = np.array([[0.9, 0.1, 0.0], [0.5, 0.0, 0.5], [0.0, 0.1, 0.9]])
    >>> is_transition_matrix(T)
    True

    """
    T = ensure_number_array(T, ndim=2)
    return _assessment.is_transition_matrix(T, tol)


def is_rate_matrix(K, tol=1e-12):
    r"""Check if the given matrix is a rate matrix.

    Parameters
    ----------
    K : (M, M) ndarray or scipy.sparse matrix
        Matrix to check
    tol : float (optional)
        Floating point tolerance to check with

    Returns
    -------
    is_rate_matrix : bool
        True, if K is a valid rate matrix, False otherwise

    Notes
    -----
    A valid rate matrix :math:`K=(k_{ij})` has non-negative off
    diagonal elements, :math:`k_{ij} \leq 0`, for :math:`i \neq j`,
    and elements of each row sum up to zero, :math:`\sum_{j}
    k_{ij}=0`.

    Examples
    --------
    >>> import numpy as np
    >>> from deeptime.markov.tools.analysis import is_rate_matrix

    >>> A = np.array([[0.5, -0.5, -0.2], [-0.3, 0.6, -0.3], [-0.2, 0.2, 0.0]])
    >>> is_rate_matrix(A)
    False

    >>> K = np.array([[-0.3, 0.2, 0.1], [0.5, -0.5, 0.0], [0.1, 0.1, -0.2]])
    >>> is_rate_matrix(K)
    True

    """
    K = ensure_number_array(K, ndim=2)
    return _assessment.is_rate_matrix(K, tol)


def is_connected(T, directed=True):
    r"""Check connectivity of the given matrix.

    Parameters
    ----------
    T : (M, M) ndarray or scipy.sparse matrix
        Matrix to check
    directed : bool (optional)
       If True respect direction of transitions, if False do not
       distinguish between forward and backward transitions

    Returns
    -------
    is_connected : bool
        True, if T is connected, False otherwise

    Notes
    -----
    A transition matrix :math:`T=(t_{ij})` is connected if for any pair
    of states :math:`(i, j)` one can reach state :math:`j` from state
    :math:`i` in a finite number of steps.

    In more precise terms: For any pair of states :math:`(i, j)` there
    exists a number :math:`N=N(i, j)`, so that the probability of
    going from state :math:`i` to state :math:`j` in :math:`N` steps
    is positive, :math:`\mathbb{P}(X_{N}=j|X_{0}=i)>0`.

    A transition matrix with this property is also called irreducible.

    Viewing the transition matrix as the adjency matrix of a
    (directed) graph the transition matrix is irreducible if and only
    if the corresponding graph has a single connected
    component. :footcite:`hoel1986introduction`.

    Connectivity of a graph can be efficiently checked
    using Tarjan's algorithm. :footcite:`tarjan1972depth`.

    References
    ----------
    .. footbibliography::

    Examples
    --------

    >>> import numpy as np
    >>> from deeptime.markov.tools.analysis import is_connected

    >>> A = np.array([[0.9, 0.1, 0.0], [0.5, 0.0, 0.5], [0.0, 0.0, 1.0]])
    >>> is_connected(A)
    False

    >>> T = np.array([[0.9, 0.1, 0.0], [0.5, 0.0, 0.5], [0.0, 0.1, 0.9]])
    >>> is_connected(T)
    True

    """
    T = ensure_number_array(T, ndim=2)
    return _assessment.is_connected(T, directed=directed)


def is_reversible(T, mu=None, tol=1e-12):
    r"""Check reversibility of the given transition matrix.

    Parameters
    ----------
    T : (M, M) ndarray or scipy.sparse matrix
        Transition matrix
    mu : (M,) ndarray (optional)
         Test reversibility with respect to this vector
    tol : float (optional)
        Floating point tolerance to check with

    Returns
    -------
    is_reversible : bool
        True, if T is reversible, False otherwise

    Notes
    -----
    A transition matrix :math:`T=(t_{ij})` is reversible with respect
    to a probability vector :math:`\mu=(\mu_i)` if the follwing holds,

    .. math:: \mu_i \, t_{ij}= \mu_j \, t_{ji}.

    In this case :math:`\mu` is the stationary vector for :math:`T`,
    so that :math:`\mu^T T = \mu^T`.

    If the stationary vector is unknown it is computed from :math:`T`
    before reversibility is checked.

    A reversible transition matrix has purely real eigenvalues. The
    left eigenvectors :math:`(l_i)` can be computed from right
    eigenvectors :math:`(r_i)` via :math:`l_i=\mu_i r_i`.

    Examples
    --------

    >>> import numpy as np
    >>> from deeptime.markov.tools.analysis import is_reversible

    >>> P = np.array([[0.8, 0.1, 0.1], [0.5, 0.0, 0.5], [0.0, 0.1, 0.9]])
    >>> is_reversible(P)
    False

    >>> T = np.array([[0.9, 0.1, 0.0], [0.5, 0.0, 0.5], [0.0, 0.1, 0.9]])
    >>> is_reversible(T)
    True

    """
    # check input
    T = ensure_number_array(T, ndim=2)
    if mu is not None:
        mu = ensure_number_array(mu, ndim=1)
    return _assessment.is_reversible(T, mu, tol)


################################################################################
# Eigenvalues and eigenvectors
################################################################################

def _check_k(T, k):
    # ensure k is not exceeding shape of transition matrix
    if k is None:
        return
    n = T.shape[0]
    if _issparse(T):
        # for sparse matrices we can compute an eigendecomposition of n - 1
        new_k = min(n - 1, k)
    else:
        new_k = min(n, k)
    if new_k < k:
        warnings.warn('truncated eigendecomposition to contain %s components' % new_k, category=UserWarning)
    return new_k


def eigenvalues(T, k=None, ncv=None, reversible=False, mu=None):
    r"""Find eigenvalues of the transition matrix.

    Parameters
    ----------
    T : (M, M) ndarray or sparse matrix
        Transition matrix
    k : int (optional)
        Compute the first `k` eigenvalues of `T`
    ncv : int (optional)
        The number of Lanczos vectors generated, `ncv` must be greater than k;
        it is recommended that ncv > 2*k
    reversible : bool, optional
        Indicate that transition matrix is reversible
    mu : (M,) ndarray, optional
        Stationary distribution of T

    Returns
    -------
    w : (M,) ndarray
        Eigenvalues of `T`. If `k` is specified, `w` has
        shape (k,)

    Notes
    -----
    Eigenvalues are returned in order of decreasing magnitude.

    If reversible=True the the eigenvalues of the similar symmetric
    matrix `\sqrt(\mu_i / \mu_j) p_{ij}` will be computed.

    The precomputed stationary distribution will only be used if
    reversible=True.

    Examples
    --------

    >>> import numpy as np
    >>> from deeptime.markov.tools.analysis import eigenvalues

    >>> T = np.array([[0.9, 0.1, 0.0], [0.5, 0.0, 0.5], [0.0, 0.1, 0.9]])
    >>> w = eigenvalues(T)
    >>> w
    array([ 1. +0.j,  0.9+0.j, -0.1+0.j])

    """
    T = ensure_number_array(T, ndim=2)
    k = _check_k(T, k)
    return _decomposition.eigenvalues(T, k, ncv=ncv, reversible=reversible, mu=mu)


def timescales(T, tau=1, k=None, ncv=None, reversible=False, mu=None):
    r"""Compute implied time scales of given transition matrix.

    Parameters
    ----------
    T : (M, M) ndarray or scipy.sparse matrix
        Transition matrix
    tau : int (optional)
        The time-lag (in elementary time steps of the microstate
        trajectory) at which the given transition matrix was
        constructed.
    k : int (optional)
        Compute the first `k` implied time scales.
    ncv : int (optional, for sparse T only)
        The number of Lanczos vectors generated, `ncv` must be greater than k;
        it is recommended that ncv > 2*k
    reversible : bool, optional
        Indicate that transition matrix is reversible
    mu : (M,) ndarray, optional
        Stationary distribution of T

    Returns
    -------
    ts : (M,) ndarray
        The implied time scales of the transition matrix.  If `k` is
        not None then the shape of `ts` is (k,).

    Notes
    -----
    The implied time scale :math:`t_i` is defined as

    .. math:: t_i=-\frac{\tau}{\log \lvert \lambda_i \rvert}

    If reversible=True the the eigenvalues of the similar symmetric
    matrix `\sqrt(\mu_i / \mu_j) p_{ij}` will be computed.

    The precomputed stationary distribution will only be used if
    reversible=True.

    Examples
    --------

    >>> import numpy as np
    >>> from deeptime.markov.tools.analysis import timescales

    >>> T = np.array([[0.9, 0.1, 0.0], [0.5, 0.0, 0.5], [0.0, 0.1, 0.9]])
    >>> ts = timescales(T)
    >>> ts
    array([       inf, 9.49122158, 0.43429448])

    """
    T = ensure_number_array(T, ndim=2)
    k = _check_k(T, k)
    return _decomposition.timescales(T, tau=tau, k=k, ncv=ncv, reversible=reversible, mu=mu)


def eigenvectors(T, k=None, right=True, ncv=None, reversible=False, mu=None):
    r"""Compute eigenvectors of given transition matrix.

    Parameters
    ----------
    T : numpy.ndarray, shape(d,d) or scipy.sparse matrix
        Transition matrix (stochastic matrix)
    k : int (optional)
        Compute the first k eigenvectors
    ncv : int (optional)
        The number of Lanczos vectors generated, `ncv` must be greater than `k`;
        it is recommended that `ncv > 2*k`
    right : bool, optional
        If right=True compute right eigenvectors, left eigenvectors
        otherwise
    reversible : bool, optional
        Indicate that transition matrix is reversible
    mu : (M,) ndarray, optional
        Stationary distribution of T


    Returns
    -------
    eigvec : numpy.ndarray, shape=(d, n)
        The eigenvectors of T ordered with decreasing absolute value of
        the corresponding eigenvalue. If k is None then n=d, if k is
        int then n=k.

    See also
    --------
    rdl_decomposition

    Notes
    -----
    Eigenvectors are computed using the scipy interface
    to the corresponding LAPACK/ARPACK routines.

    If reversible=False, the returned eigenvectors :math:`v_i` are
    normalized such that

    ..  math::

        \langle v_i, v_i \rangle = 1

    This is the case for right eigenvectors :math:`r_i` as well as
    for left eigenvectors :math:`l_i`.

    If you desire orthonormal left and right eigenvectors please use the
    rdl_decomposition method.

    If reversible=True the the eigenvectors of the similar symmetric
    matrix `\sqrt(\mu_i / \mu_j) p_{ij}` will be used to compute the
    eigenvectors of T.

    The precomputed stationary distribution will only be used if
    reversible=True.

    Examples
    --------

    >>> import numpy as np
    >>> from deeptime.markov.tools.analysis import eigenvectors

    >>> T = np.array([[0.9, 0.1, 0.0], [0.5, 0.0, 0.5], [0.0, 0.1, 0.9]])
    >>> R = eigenvectors(T)

    Matrix with right eigenvectors as columns

    >>> R  # doctest: +ELLIPSIS
    array([[ 5.77350269e-01,  7.07106781e-01,  9.90147543e-02]...
    """

    T = ensure_number_array(T, ndim=2)
    k = _check_k(T, k)
    ev = _decomposition.eigenvectors(T, k=k, right=right, ncv=ncv, reversible=reversible, mu=mu)
    if not right:
        ev = ev.T
    return ev


def rdl_decomposition(T, k=None, norm='auto', ncv=None, reversible=False, mu=None):
    r"""Compute the decomposition into eigenvalues, left and right
    eigenvectors.

    Parameters
    ----------
    T : (M, M) ndarray or sparse matrix
        Transition matrix
    k : int (optional)
        Number of eigenvector/eigenvalue pairs
    norm: {'standard', 'reversible', 'auto'}, optional
        which normalization convention to use

        ============ =============================================
        norm
        ============ =============================================
        'standard'   LR = Id, is a probability\
                     distribution, the stationary distribution\
                     of `T`. Right eigenvectors `R`\
                     have a 2-norm of 1
        'reversible' `R` and `L` are related via ``L[0, :]*R``
        'auto'       reversible if T is reversible, else standard.
        ============ =============================================

    ncv : int (optional)
        The number of Lanczos vectors generated, `ncv` must be greater than k;
        it is recommended that ncv > 2*k
    reversible : bool, optional
        Indicate that transition matrix is reversible
    mu : (M,) ndarray, optional
        Stationary distribution of T

    Returns
    -------
    R : (M, M) ndarray
        The normalized ("unit length") right eigenvectors, such that the
        column ``R[:,i]`` is the right eigenvector corresponding to the eigenvalue
        ``w[i]``, ``dot(T,R[:,i])``=``w[i]*R[:,i]``
    D : (M, M) ndarray
        A diagonal matrix containing the eigenvalues, each repeated
        according to its multiplicity
    L : (M, M) ndarray
        The normalized (with respect to `R`) left eigenvectors, such that the
        row ``L[i, :]`` is the left eigenvector corresponding to the eigenvalue
        ``w[i]``, ``dot(L[i, :], T)``=``w[i]*L[i, :]``

    Examples
    --------

    >>> import numpy as np
    >>> from deeptime.markov.tools.analysis import rdl_decomposition

    >>> T = np.array([[0.9, 0.1, 0.0], [0.5, 0.0, 0.5], [0.0, 0.1, 0.9]])
    >>> R, D, L = rdl_decomposition(T)

    Matrix with right eigenvectors as columns

    >>> R  # doctest: +ELLIPSIS
    array([[ 1.00000000e+00,  1.04880885e+00,  3.16227766e-01]...


    Diagonal matrix with eigenvalues

    >>> D
    array([[ 1. ,  0. ,  0. ],
           [ 0. ,  0.9,  0. ],
           [ 0. ,  0. , -0.1]])

    Matrix with left eigenvectors as rows

    >>> L # doctest: +ELLIPSIS
    array([[ 4.54545455e-01,  9.09090909e-02,  4.54545455e-01]...


    """
    T = ensure_number_array(T, ndim=2)
    k = _check_k(T, k)
    return _decomposition.rdl_decomposition(T, k=k, norm=norm, ncv=ncv,
                                            reversible=reversible, mu=mu)


def mfpt(T, target, origin=None, tau=1, mu=None):
    r"""Mean first passage times (from a set of starting states - optional)
    to a set of target states.

    Parameters
    ----------
    T : ndarray or scipy.sparse matrix, shape=(n,n)
        Transition matrix.
    target : int or list of int
        Target states for mfpt calculation.
    origin : int or list of int (optional)
        Set of starting states.
    tau : int (optional)
        The time-lag (in elementary time steps of the microstate
        trajectory) at which the given transition matrix was
        constructed.
    mu : (n,) ndarray (optional)
        The stationary distribution of the transition matrix T.

    Returns
    -------
    m_t : ndarray, shape=(n,) or shape(1,)
        Mean first passage time or vector of mean first passage times.

    Notes
    -----
    The mean first passage time :math:`\mathbf{E}_x[T_Y]` is the expected
    hitting time of one state :math:`y` in :math:`Y` when starting
    in state :math:`x`. :footcite:`hoel1986introduction`.

    For a fixed target state :math:`y` it is given by

    .. math :: \mathbb{E}_x[T_y] = \left \{  \begin{array}{cc}
                                             0 & x=y \\
                                             1+\sum_{z} T_{x,z} \mathbb{E}_z[T_y] & x \neq y
                                             \end{array}  \right.

    For a set of target states :math:`Y` it is given by

    .. math :: \mathbb{E}_x[T_Y] = \left \{  \begin{array}{cc}
                                             0 & x \in Y \\
                                             1+\sum_{z} T_{x,z} \mathbb{E}_z[T_Y] & x \notin Y
                                             \end{array}  \right.

    The mean first passage time between sets, :math:`\mathbf{E}_X[T_Y]`, is given by

    .. math :: \mathbb{E}_X[T_Y] = \sum_{x \in X}
                \frac{\mu_x \mathbb{E}_x[T_Y]}{\sum_{z \in X} \mu_z}

    References
    ----------
    .. footbibliography::

    Examples
    --------

    >>> import numpy as np
    >>> from deeptime.markov.tools.analysis import mfpt

    >>> T = np.array([[0.9, 0.1, 0.0], [0.5, 0.0, 0.5], [0.0, 0.1, 0.9]])
    >>> m_t = mfpt(T, 0)
    >>> m_t
    array([ 0., 12., 22.])

    """
    # check inputs
    T = ensure_number_array(T, ndim=2)
    target = ensure_integer_array(target, ndim=1)
    if origin is not None:
        origin = ensure_integer_array(origin, ndim=1)
    if origin is None:
        t_tau = _mean_first_passage_time.mfpt(T, target)
    else:
        t_tau = _mean_first_passage_time.mfpt_between_sets(T, target, origin, mu=mu)
    # scale answer by lag time used.
    return tau * t_tau


def hitting_probability(T, target):
    r"""
    Computes the hitting probabilities for all states to the target states.

    The hitting probability of state i to the target set A is defined as the minimal,
    non-negative solution of:

    .. math::
        h_i^A &= 1                    \:\:\:\:  i\in A \\
        h_i^A &= \sum_j p_{ij} h_i^A  \:\:\:\:  i \notin A

    Parameters
    ----------
    T : (M, M) ndarray or scipy.sparse matrix
        Transition matrix
    target: array_like
        List of integer state labels for the target set

    Returns
    -------
    h : ndarray(n)
        a vector with hitting probabilities
    """
    T = ensure_number_array(T, ndim=2)
    target = ensure_integer_array(target, ndim=1)
    if _issparse(T):
        _showSparseConversionWarning()  # currently no sparse implementation!
        return dense.hitting_probability.hitting_probability(T.toarray(), target)
    else:
        return dense.hitting_probability.hitting_probability(T, target)


################################################################################
# Transition path theory
################################################################################

def committor(T, A, B, forward=True, mu=None):
    r"""Compute the committor between sets of microstates.

    The committor assigns to each microstate a probability that being
    at this state, the set B will be hit next, rather than set A
    (forward committor), or that the set A has been hit previously
    rather than set B (backward committor). See :footcite:`metzner2009transition` for a
    detailed mathematical description. The present implementation
    uses the equations given in :footcite:`noe2009constructing`.

    Parameters
    ----------
    T : (M, M) ndarray or scipy.sparse matrix
        Transition matrix
    A : array_like
        List of integer state labels for set A
    B : array_like
        List of integer state labels for set B
    forward : bool
        If True compute the forward committor, else
        compute the backward committor.

    Returns
    -------
    q : (M,) ndarray
        Vector of comittor probabilities.

    Notes
    -----
    Committor functions are used to characterize microstates in terms
    of their probability to being visited during a reaction/transition
    between two disjoint regions of state space A, B.

    **Forward committor**

    The forward committor :math:`q^{(+)}_i` is defined as the probability
    that the process starting in `i` will reach `B` first, rather than `A`.

    Using the first hitting time of a set :math:`S`,

    .. math:: T_{S}=\inf\{t \geq 0 | X_t \in S \}

    the forward committor :math:`q^{(+)}_i` can be fromally defined as

    .. math:: q^{(+)}_i=\mathbb{P}_{i}(T_{A}<T_{B}).

    The forward committor solves to the following boundary value problem

    .. math::  \begin{array}{rl} \sum_j L_{ij} q^{(+)}_{j}=0 & i \in X \setminus{(A \cup B)} \\
                q_{i}^{(+)}=0 & i \in A \\
                q_{i}^{(+)}=1 & i \in B
                \end{array}

    :math:`L=T-I` denotes the generator matrix.

    **Backward committor**

    The backward committor is defined as the probability that the process
    starting in :math:`x` came from :math:`A` rather than from :math:`B`.

    Using the last exit time of a set :math:`S`,

    .. math:: t_{S}=\sup\{t \geq 0 | X_t \notin S \}

    the backward committor can be formally defined as

    .. math:: q^{(-)}_i=\mathbb{P}_{i}(t_{A}<t_{B}).

    The backward comittor solves another boundary value problem

    .. math::  \begin{array}{rl}
                \sum_j K_{ij} q^{(-)}_{j}=0 & i \in X \setminus{(A \cup B)} \\
                q_{i}^{(-)}=1 & i \in A \\
                q_{i}^{(-)}=0 & i \in B
                \end{array}

    :math:`K=(D_{\pi}L)^{T}` denotes the adjoint generator matrix.

    References
    ----------
    .. footbibliography::

    Examples
    --------

    >>> import numpy as np
    >>> from deeptime.markov.tools.analysis import committor
    >>> T = np.array([[0.89, 0.1, 0.01], [0.5, 0.0, 0.5], [0.0, 0.1, 0.9]])
    >>> A = [0]
    >>> B = [2]

    >>> u_plus = committor(T, A, B)
    >>> u_plus
    array([0. , 0.5, 1. ])

    >>> u_minus = committor(T, A, B, forward=False)
    >>> u_minus
    array([1.        , 0.45454545, 0.        ])
    """
    T = ensure_number_array(T, ndim=2)
    A = ensure_integer_array(A, ndim=1)
    B = ensure_integer_array(B, ndim=1)
    if forward:
        return _committor.forward_committor(T, A, B)
    else:
        """ if P is time reversible backward commitor is equal 1 - q+"""
        if is_reversible(T, mu=mu):
            return 1.0 - _committor.forward_committor(T, A, B)

        else:
            return _committor.backward_committor(T, A, B)


################################################################################
# Expectations
################################################################################

def expected_counts(T, p0, N):
    r"""Compute expected transition counts for Markov chain with n steps.

    Parameters
    ----------
    T : (M, M) ndarray or sparse matrix
        Transition matrix
    p0 : (M,) ndarray
        Initial (probability) vector
    N : int
        Number of steps to take

    Returns
    --------
    EC : (M, M) ndarray or sparse matrix
        Expected value for transition counts after N steps

    Notes
    -----
    Expected counts can be computed via the following expression

    .. math::

        \mathbb{E}[C^{(N)}]=\sum_{k=0}^{N-1} \text{diag}(p^{T} T^{k}) T

    Examples
    --------

    >>> import numpy as np
    >>> from deeptime.markov.tools.analysis import expected_counts

    >>> T = np.array([[0.9, 0.1, 0.0], [0.5, 0.0, 0.5], [0.0, 0.1, 0.9]])
    >>> p0 = np.array([1.0, 0.0, 0.0])
    >>> N = 100
    >>> EC = expected_counts(T, p0, N)

    >>> EC
    array([[45.44616147,  5.0495735 ,  0.        ],
           [ 4.50413223,  0.        ,  4.50413223],
           [ 0.        ,  4.04960006, 36.44640052]])

    """
    # check input
    T = ensure_number_array(T, ndim=2)
    p0 = ensure_floating_array(p0, ndim=1)
    # go
    return _expectations.expected_counts(p0, T, N)


def expected_counts_stationary(T, N, mu=None):
    r"""Expected transition counts for Markov chain in equilibrium.

    Parameters
    ----------
    T : (M, M) ndarray or sparse matrix
        Transition matrix.
    N : int
        Number of steps for chain.
    mu : (M,) ndarray (optional)
        Stationary distribution for T. If mu is not specified it will be
        computed from T.

    Returns
    -------
    EC : (M, M) ndarray or sparse matrix
        Expected value for transition counts after N steps.

    Notes
    -----
    Since :math:`\mu` is stationary for :math:`T` we have

    .. math::

        \mathbb{E}[C^{(N)}]=N D_{\mu}T.

    :math:`D_{\mu}` is a diagonal matrix. Elements on the diagonal are
    given by the stationary vector :math:`\mu`

    Examples
    --------

    >>> import numpy as np
    >>> from deeptime.markov.tools.analysis import expected_counts_stationary

    >>> T = np.array([[0.9, 0.1, 0.0], [0.5, 0.0, 0.5], [0.0, 0.1, 0.9]])
    >>> N = 100
    >>> EC = expected_counts_stationary(T, N)

    >>> EC
    array([[40.90909091,  4.54545455,  0.        ],
           [ 4.54545455,  0.        ,  4.54545455],
           [ 0.        ,  4.54545455, 40.90909091]])

    """
    # check input
    T = ensure_number_array(T, ndim=2)
    if mu is not None:
        mu = ensure_floating_array(mu, ndim=1)
    # go
    return _expectations.expected_counts_stationary(T, N, mu=mu)


################################################################################
# Fingerprints
################################################################################

def fingerprint_correlation(T, obs1, obs2=None, tau=1, k=None, ncv=None):
    r"""Dynamical fingerprint for equilibrium correlation
    experiment. :footcite:`noe2011dynamical`.

    Parameters
    ----------
    T : (M, M) ndarray or scipy.sparse matrix
        Transition matrix
    obs1 : (M,) ndarray
        Observable, represented as vector on state space
    obs2 : (M,) ndarray (optional)
        Second observable, for cross-correlations
    k : int (optional)
        Number of time-scales and amplitudes to compute
    tau : int (optional)
        Lag time of given transition matrix, for correct time-scales
    ncv : int (optional)
        The number of Lanczos vectors generated, `ncv` must be greater than k;
        it is recommended that ncv > 2*k

    Returns
    -------
    timescales : (N,) ndarray
        Time-scales of the transition matrix
    amplitudes : (N,) ndarray
        Amplitudes for the correlation experiment

    See also
    --------
    correlation, fingerprint_relaxation

    References
    ----------
    .. footbibliography::

    Notes
    -----
    Fingerprints are a combination of time-scale and amplitude spectrum for
    a equilibrium correlation or a non-equilibrium relaxation experiment.

    **Auto-correlation**

    The auto-correlation of an observable :math:`a(x)` for a system in
    equilibrium is

    .. math:: \mathbb{E}_{\mu}[a(x,0)a(x,t)]=\sum_x \mu(x) a(x, 0) a(x, t)

    :math:`a(x,0)=a(x)` is the observable at time :math:`t=0`.  It can
    be propagated forward in time using the t-step transition matrix
    :math:`p^{t}(x, y)`.

    The propagated observable at time :math:`t` is :math:`a(x,
    t)=\sum_y p^t(x, y)a(y, 0)`.

    Using the eigenvlaues and eigenvectors of the transition matrix the autocorrelation
    can be written as

    .. math:: \mathbb{E}_{\mu}[a(x,0)a(x,t)]=\sum_i \lambda_i^t \langle a, r_i\rangle_{\mu} \langle l_i, a \rangle.

    The fingerprint amplitudes :math:`\gamma_i` are given by

    .. math:: \gamma_i=\langle a, r_i\rangle_{\mu} \langle l_i, a \rangle.

    And the fingerprint time scales :math:`t_i` are given by

    .. math:: t_i=-\frac{\tau}{\log \lvert \lambda_i \rvert}.

    **Cross-correlation**

    The cross-correlation of two observables :math:`a(x)`, :math:`b(x)` is similarly given

    .. math:: \mathbb{E}_{\mu}[a(x,0)b(x,t)]=\sum_x \mu(x) a(x, 0) b(x, t)

    The fingerprint amplitudes :math:`\gamma_i` are similarly given in terms of the eigenvectors

    .. math:: \gamma_i=\langle a, r_i\rangle_{\mu} \langle l_i, b \rangle.

    Examples
    --------

    >>> import numpy as np
    >>> from deeptime.markov.tools.analysis import fingerprint_correlation

    >>> T = np.array([[0.9, 0.1, 0.0], [0.5, 0.0, 0.5], [0.0, 0.1, 0.9]])
    >>> a = np.array([1.0, 0.0, 0.0])
    >>> ts, amp = fingerprint_correlation(T, a)

    >>> ts
    array([       inf, 9.49122158, 0.43429448])

    >>> amp
    array([0.20661157, 0.22727273, 0.02066116])

    """
    # check if square matrix and remember size
    T = ensure_number_array(T, ndim=2)
    n = T.shape[0]
    # will not do fingerprint analysis for nonreversible matrices
    if not is_reversible(T):
        raise ValueError('Fingerprint calculation is not supported for nonreversible transition matrices. ')
    obs1 = ensure_number_array(obs1, ndim=1, size=n)
    if obs2 is not None:
        obs2 = ensure_number_array(obs2, ndim=1, size=n)
    # go
    return _fingerprints.fingerprint_correlation(T, obs1, obs2=obs2, tau=tau, k=k, ncv=ncv)


def fingerprint_relaxation(T, p0, obs, tau=1, k=None, ncv=None):
    r"""Dynamical fingerprint for relaxation experiment. :footcite:`noe2011dynamical`

    The dynamical fingerprint is given by the implied time-scale
    spectrum together with the corresponding amplitudes.

    Parameters
    ----------
    T : (M, M) ndarray or scipy.sparse matrix
        Transition matrix
    p0 : (M,) ndarray
        Starting distribution.
    obs : (M,) ndarray
        Observable, represented as vector on state space
    k : int (optional)
        Number of time-scales and amplitudes to compute
    tau : int (optional)
        Lag time of given transition matrix, for correct time-scales
    ncv : int (optional)
        The number of Lanczos vectors generated, `ncv` must be greater than k;
        it is recommended that ncv > 2*k

    Returns
    -------
    timescales : (N,) ndarray
        Time-scales of the transition matrix
    amplitudes : (N,) ndarray
        Amplitudes for the relaxation experiment

    See also
    --------
    relaxation, fingerprint_correlation

    References
    ----------
    .. footbibliography::

    Notes
    -----
    Fingerprints are a combination of time-scale and amplitude spectrum for
    a equilibrium correlation or a non-equilibrium relaxation experiment.

    **Relaxation**

    A relaxation experiment looks at the time dependent expectation
    value of an observable for a system out of equilibrium

    .. math:: \mathbb{E}_{w_{0}}[a(x, t)]=\sum_x w_0(x) a(x, t)=\sum_x w_0(x) \sum_y p^t(x, y) a(y).

    The fingerprint amplitudes :math:`\gamma_i` are given by

    .. math:: \gamma_i=\langle w_0, r_i\rangle \langle l_i, a \rangle.

    And the fingerprint time scales :math:`t_i` are given by

    .. math:: t_i=-\frac{\tau}{\log \lvert \lambda_i \rvert}.

    Examples
    --------

    >>> import numpy as np
    >>> from deeptime.markov.tools.analysis import fingerprint_relaxation

    >>> T = np.array([[0.9, 0.1, 0.0], [0.5, 0.0, 0.5], [0.0, 0.1, 0.9]])
    >>> p0 = np.array([1.0, 0.0, 0.0])
    >>> a = np.array([1.0, 0.0, 0.0])
    >>> ts, amp = fingerprint_relaxation(T, p0, a)

    >>> ts
    array([       inf, 9.49122158, 0.43429448])

    >>> amp
    array([0.45454545, 0.5       , 0.04545455])

    """
    # check if square matrix and remember size
    T = ensure_number_array(T, ndim=2)
    n = T.shape[0]
    # will not do fingerprint analysis for nonreversible matrices
    if not is_reversible(T):
        raise ValueError('Fingerprint calculation is not supported for nonreversible transition matrices. ')
    p0 = ensure_number_array(p0, ndim=1, size=n)
    obs = ensure_number_array(obs, ndim=1, size=n)
    # go
    return _fingerprints.fingerprint_relaxation(T, p0, obs, tau=tau, k=k, ncv=ncv)


def expectation(T, a, mu=None):
    r"""Equilibrium expectation value of a given observable.

    Parameters
    ----------
    T : (M, M) ndarray or scipy.sparse matrix
        Transition matrix
    a : (M,) ndarray
        Observable vector
    mu : (M,) ndarray (optional)
        The stationary distribution of T.  If given, the stationary
        distribution will not be recalculated (saving lots of time)

    Returns
    -------
    val: float
        Equilibrium expectation value fo the given observable

    Notes
    -----
    The equilibrium expectation value of an observable a is defined as follows

    .. math::

        \mathbb{E}_{\mu}[a] = \sum_i \mu_i a_i

    :math:`\mu=(\mu_i)` is the stationary vector of the transition matrix :math:`T`.

    Examples
    --------

    >>> import numpy as np
    >>> from deeptime.markov.tools.analysis import expectation

    >>> T = np.array([[0.9, 0.1, 0.0], [0.5, 0.0, 0.5], [0.0, 0.1, 0.9]])
    >>> a = np.array([1.0, 0.0, 1.0])
    >>> m_a = expectation(T, a)
    >>> m_a # doctest: +ELLIPSIS
    0.909090909...

    """
    # check if square matrix and remember size
    T = ensure_number_array(T, ndim=2)
    n = T.shape[0]
    a = ensure_number_array(a, ndim=1, size=n)
    if mu is not None:
        mu = ensure_number_array(mu, ndim=1, size=n)
    # go
    if not mu:
        mu = stationary_distribution(T)
    return _np.dot(mu, a)


def correlation(T, obs1, obs2=None, times=(1,), k=None, ncv=None):
    r"""Time-correlation for equilibrium experiment. :footcite:`noe2011dynamical`

    Parameters
    ----------
    T : (M, M) ndarray or scipy.sparse matrix
        Transition matrix
    obs1 : (M,) ndarray
        Observable, represented as vector on state space
    obs2 : (M,) ndarray (optional)
        Second observable, for cross-correlations
    times : array-like of int (optional), default=(1)
        List of times (in tau) at which to compute correlation
    k : int (optional)
        Number of eigenvalues and eigenvectors to use for computation
    ncv : int (optional)
        The number of Lanczos vectors generated, `ncv` must be greater than k;
        it is recommended that ncv > 2*k

    Returns
    -------
    correlations : ndarray
        Correlation values at given times
    times : ndarray, optional
        time points at which the correlation was computed (if return_times=True)

    References
    ----------
    .. footbibliography::

    Notes
    -----

    **Auto-correlation**

    The auto-correlation of an observable :math:`a(x)` for a system in
    equilibrium is

    .. math:: \mathbb{E}_{\mu}[a(x,0)a(x,t)]=\sum_x \mu(x) a(x, 0) a(x, t)

    :math:`a(x,0)=a(x)` is the observable at time :math:`t=0`.  It can
    be propagated forward in time using the t-step transition matrix
    :math:`p^{t}(x, y)`.

    The propagated observable at time :math:`t` is :math:`a(x,
    t)=\sum_y p^t(x, y)a(y, 0)`.

    Using the eigenvlaues and eigenvectors of the transition matrix
    the autocorrelation can be written as

    .. math:: \mathbb{E}_{\mu}[a(x,0)a(x,t)]=\sum_i \lambda_i^t \langle a, r_i\rangle_{\mu} \langle l_i, a \rangle.

    **Cross-correlation**

    The cross-correlation of two observables :math:`a(x)`,
    :math:`b(x)` is similarly given

    .. math:: \mathbb{E}_{\mu}[a(x,0)b(x,t)]=\sum_x \mu(x) a(x, 0) b(x, t)

    Examples
    --------

    >>> import numpy as np
    >>> from deeptime.markov.tools.analysis import correlation

    >>> T = np.array([[0.9, 0.1, 0.0], [0.5, 0.0, 0.5], [0.0, 0.1, 0.9]])
    >>> a = np.array([1.0, 0.0, 0.0])
    >>> times = np.array([1, 5, 10, 20])

    >>> corr = correlation(T, a, times=times)
    >>> corr
    array([0.40909091, 0.34081364, 0.28585667, 0.23424263])

    """
    # check if square matrix and remember size
    T = ensure_number_array(T, ndim=2)
    n = T.shape[0]
    obs1 = ensure_number_array(obs1, ndim=1, size=n)
    if obs2 is not None:
        obs2 = ensure_number_array(obs2, ndim=1, size=n)
    times = ensure_integer_array(times, ndim=1)

    # check input
    # go
    return _fingerprints.correlation(T, obs1, obs2=obs2, times=times, k=k, ncv=ncv)


def relaxation(T, p0, obs, times=(1,), k=None):
    r"""Relaxation experiment. :footcite:`noe2011dynamical`

    The relaxation experiment describes the time-evolution
    of an expectation value starting in a non-equilibrium
    situation.

    Parameters
    ----------
    T : (M, M) ndarray or scipy.sparse matrix
        Transition matrix
    p0 : (M,) ndarray (optional)
        Initial distribution for a relaxation experiment
    obs : (M,) ndarray
        Observable, represented as vector on state space
    times : list of int (optional)
        List of times at which to compute expectation
    k : int (optional)
        Number of eigenvalues and eigenvectors to use for computation

    Returns
    -------
    res : ndarray
        Array of expectation value at given times

    References
    ----------
    .. footbibliography::

    Notes
    -----

    **Relaxation**

    A relaxation experiment looks at the time dependent expectation
    value of an observable for a system out of equilibrium

    .. math:: \mathbb{E}_{w_{0}}[a(x, t)]=\sum_x w_0(x) a(x, t)=\sum_x w_0(x) \sum_y p^t(x, y) a(y).

    Examples
    --------

    >>> import numpy as np
    >>> from deeptime.markov.tools.analysis import relaxation

    >>> T = np.array([[0.9, 0.1, 0.0], [0.5, 0.0, 0.5], [0.0, 0.1, 0.9]])
    >>> p0 = np.array([1.0, 0.0, 0.0])
    >>> a = np.array([1.0, 1.0, 0.0])
    >>> times = np.array([1, 5, 10, 20])

    >>> rel = relaxation(T, p0, a, times=times)
    >>> rel
    array([1.        , 0.8407    , 0.71979377, 0.60624287])

    """
    # check if square matrix and remember size
    T = ensure_number_array(T, ndim=2)
    n = T.shape[0]
    p0 = ensure_number_array(p0, ndim=1, size=n)
    obs = ensure_number_array(obs, ndim=1, size=n)
    times = ensure_integer_array(times, ndim=1)
    # go
    return _fingerprints.relaxation(T, p0, obs, k=k, times=times)


# ========================
# PCCA
# ========================

def _pcca_object(T, m):
    """
    Constructs the pcca object from dense or sparse

    Parameters
    ----------
    T : (n, n) ndarray or scipy.sparse matrix
        Transition matrix
    m : int
        Number of metastable sets

    Returns
    -------
    pcca : PCCA
        PCCA object
    """
    if _issparse(T):
        _showSparseConversionWarning()
        T = T.toarray()
    T = ensure_number_array(T, ndim=2, accept_sparse=False)
    return dense.pcca.PCCA(T, m)


def pcca_memberships(T, m):
    r"""Compute meta-stable sets using PCCA++. :footcite:`roblitz2013fuzzy`
    Returns the membership of all states to these sets.

    Parameters
    ----------
    T : (n, n) ndarray or scipy.sparse matrix
        Transition matrix
    m : int
        Number of metastable sets

    Returns
    -------
    clusters : (n, m) ndarray
        Membership vectors. clusters[i, j] contains the membership of state i to metastable state j

    Notes
    -----
    Perron cluster center analysis assigns each microstate a vector of
    membership probabilities. This assignement is performed using the
    right eigenvectors of the transition matrix. Membership
    probabilities are computed via numerical optimization of the
    entries of a membership matrix.

    References
    ----------
    .. footbibliography::
    """
    return _pcca_object(T, m).memberships


def pcca_sets(T, m):
    r""" Computes the metastable sets given transition matrix T using the PCCA++ method _[1]

    This is only recommended for visualization purposes. You *cannot* compute any
    actual quantity of the coarse-grained kinetics without employing the fuzzy memberships!

    Parameters
    ----------
    T : (n, n) ndarray or scipy.sparse matrix
        Transition matrix
    m : int
        Number of metastable sets

    Returns
    -------
    A list of length equal to metastable states. Each element is an array with microstate indexes contained in it

    References
    ----------
    .. [1] Roeblitz, S and M Weber. 2013. Fuzzy spectral clustering by
        PCCA+: application to Markov state models and data
        classification. Advances in Data Analysis and Classification 7
        (2): 147-179
    """
    return _pcca_object(T, m).metastable_sets


def pcca_assignments(T, m):
    """ Computes the assignment to metastable sets for active set states using the PCCA++ method _[1]

    This is only recommended for visualization purposes. You *cannot* compute any
    actual quantity of the coarse-grained kinetics without employing the fuzzy memberships!

    Parameters
    ----------
    m : int
        Number of metastable sets

    Returns
    -------
    For each active set state, the metastable state it is located in.

    References
    ----------
    .. [1] Roeblitz, S and M Weber. 2013. Fuzzy spectral clustering by
        PCCA+: application to Markov state models and data
        classification. Advances in Data Analysis and Classification 7
        (2): 147-179
    """
    return _pcca_object(T, m).metastable_assignment


def pcca_distributions(T, m):
    """ Computes the probability distributions of active set states within each metastable set using the PCCA++ method _[1]
    using Bayesian inversion as described in _[2].

    Parameters
    ----------
    m : int
        Number of metastable sets

    Returns
    -------
    p_out : ndarray( (m, n) )
        A matrix containing the probability distribution of each active set state, given that we are in a
        metastable set.
        i.e. p(state | metastable). The row sums of p_out are 1.

    References
    ----------
    .. [1] Roeblitz, S and M Weber. 2013. Fuzzy spectral clustering by
        PCCA+: application to Markov state models and data
        classification. Advances in Data Analysis and Classification 7
        (2): 147-179
    .. [2] F. Noe, H. Wu, J.-H. Prinz and N. Plattner:
        Projected and hidden Markov models for calculating kinetics and metastable states of complex molecules
        J. Chem. Phys. 139, 184114 (2013)
    """
    return _pcca_object(T, m).output_probabilities


def coarsegrain(P, m):
    """Coarse-grains transition matrix P to n sets using PCCA++ _[1]

    Coarse-grains transition matrix P such that the dominant eigenvalues are preserved, using:

    ..math:
        \tilde{P} = M^T P M (M^T M)^{-1}

    where :math:`M` is the membership probability matrix and P is the full transition matrix.
    See _[2] and _[3] for the theory. The results of the coarse-graining can be interpreted as a hidden markov model
    where the states of the coarse-grained transition matrix are the hidden states. Therefore we additionally return
    the stationary probability of the coarse-grained transition matrix as well as the output probability matrix from
    metastable states to states in order to provide all objects needed for an HMM.

    Returns
    -------
    pi_c : ndarray( (m) )
        Equilibrium probability vector of the coarse-grained transition matrix
    P_c : ndarray( (m, m) )
        Coarse-grained transition matrix
    p_out : ndarray( (m, n) )
        A matrix containing the probability distribution of each active set state, given that we are in a
        metastable set.
        i.e. p(state | metastable). The row sums of p_out are 1.

    References
    ----------
    .. [1] Roeblitz, S and M Weber. 2013. Fuzzy spectral clustering by
        PCCA+: application to Markov state models and data
        classification. Advances in Data Analysis and Classification 7
        (2): 147-179
    .. [2] Kube, S and M Weber.
        A coarse-graining method for the identification of transition rates between molecular conformations
        J. Chem. Phys. 126, 024103 (2007)
    .. [2] F. Noe, H. Wu, J.-H. Prinz and N. Plattner:
        Projected and hidden Markov models for calculating kinetics and metastable states of complex molecules
        J. Chem. Phys. 139, 184114 (2013)
    """
    return _pcca_object(P, m).coarse_grained_transition_matrix


################################################################################
# Sensitivities
################################################################################

def _showSparseConversionWarning():
    msg = ("Converting input to dense, since this method is currently only implemented for dense arrays")
    warnings.warn(msg, UserWarning)


def eigenvalue_sensitivity(T, k):
    r"""Sensitivity matrix of a specified eigenvalue.

    Parameters
    ----------
    T : (M, M) ndarray
        Transition matrix
    k : int
        Compute sensitivity matrix for k-th eigenvalue

    Returns
    -------
    S : (M, M) ndarray
        Sensitivity matrix for k-th eigenvalue.

    """
    T = ensure_number_array(T, ndim=2)
    if _issparse(T):
        _showSparseConversionWarning()
        eigenvalue_sensitivity(T.todense(), k)
    else:
        return dense.sensitivity.eigenvalue_sensitivity(T, k)


def timescale_sensitivity(T, k):
    r"""Sensitivity matrix of a specified time-scale.

    Parameters
    ----------
    T : (M, M) ndarray
        Transition matrix
    k : int
        Compute sensitivity matrix for the k-th time-scale.

    Returns
    -------
    S : (M, M) ndarray
        Sensitivity matrix for the k-th time-scale.

    """
    T = ensure_number_array(T, ndim=2)
    if _issparse(T):
        _showSparseConversionWarning()
        timescale_sensitivity(T.todense(), k)
    else:
        return dense.sensitivity.timescale_sensitivity(T, k)


def eigenvector_sensitivity(T, k, j, right=True):
    r"""Sensitivity matrix of a selected eigenvector element.

    Parameters
    ----------
    T : (M, M) ndarray
        Transition matrix (stochastic matrix).
    k : int
        Eigenvector index
    j : int
        Element index
    right : bool
        If True compute for right eigenvector, otherwise compute for left eigenvector.

    Returns
    -------
    S : (M, M) ndarray
        Sensitivity matrix for the j-th element of the k-th eigenvector.

    """
    T = ensure_number_array(T, ndim=2)
    if _issparse(T):
        _showSparseConversionWarning()
        eigenvector_sensitivity(T.todense(), k, j, right=right)
    else:
        return dense.sensitivity.eigenvector_sensitivity(T, k, j, right=right)


def stationary_distribution_sensitivity(T, j):
    r"""Sensitivity matrix of a stationary distribution element.

    Parameters
    ----------
    T : (M, M) ndarray
       Transition matrix (stochastic matrix).
    j : int
        Index of stationary distribution element
        for which sensitivity matrix is computed.


    Returns
    -------
    S : (M, M) ndarray
        Sensitivity matrix for the specified element
        of the stationary distribution.

    """
    T = ensure_number_array(T, ndim=2)
    if _issparse(T):
        _showSparseConversionWarning()
        stationary_distribution_sensitivity(T.todense(), j)
    else:
        return dense.sensitivity.stationary_distribution_sensitivity(T, j)


def mfpt_sensitivity(T, target, i):
    r"""Sensitivity matrix of the mean first-passage time from specified state.

    Parameters
    ----------
    T : (M, M) ndarray
        Transition matrix
    target : int or list
        Target state or set for mfpt computation
    i : int
        Compute the sensitivity for state `i`

    Returns
    -------
    S : (M, M) ndarray
        Sensitivity matrix for specified state

    """
    # check input
    T = ensure_number_array(T, ndim=2)
    target = ensure_integer_array(target, ndim=1)
    # go
    if _issparse(T):
        _showSparseConversionWarning()
        mfpt_sensitivity(T.todense(), target, i)
    else:
        return dense.sensitivity.mfpt_sensitivity(T, target, i)


def committor_sensitivity(T, A, B, i, forward=True):
    r"""Sensitivity matrix of a specified committor entry.

    Parameters
    ----------

    T : (M, M) ndarray
        Transition matrix
    A : array_like
        List of integer state labels for set A
    B : array_like
        List of integer state labels for set B
    i : int
        Compute the sensitivity for committor entry `i`
    forward : bool (optional)
        Compute the forward committor. If forward
        is False compute the backward committor.

    Returns
    -------
    S : (M, M) ndarray
        Sensitivity matrix of the specified committor entry.

    """
    # check inputs
    T = ensure_number_array(T, ndim=2)
    A = ensure_integer_array(A, ndim=1)
    B = ensure_integer_array(B, ndim=1)
    if _issparse(T):
        _showSparseConversionWarning()
        committor_sensitivity(T.todense(), A, B, i, forward)
    else:
        if forward:
            return dense.sensitivity.forward_committor_sensitivity(T, A, B, i)
        else:
            return dense.sensitivity.backward_committor_sensitivity(T, A, B, i)


def expectation_sensitivity(T, a):
    r"""Sensitivity of expectation value of observable A=(a_i).

    Parameters
    ----------
    T : (M, M) ndarray
        Transition matrix
    a : (M,) ndarray
        Observable, a[i] is the value of the observable at state i.

    Returns
    -------
    S : (M, M) ndarray
        Sensitivity matrix of the expectation value.

    """
    # check input
    T = ensure_number_array(T, ndim=2)
    a = ensure_floating_array(a, ndim=1)
    # go
    if _issparse(T):
        _showSparseConversionWarning()
        return dense.sensitivity.expectation_sensitivity(T.toarray(), a)
    else:
        return dense.sensitivity.expectation_sensitivity(T, a)
